=head1 NAME

Gdraw2::AbstractTrack - a series of genomic features

=head1 SYNOPSIS

 # This is only a abstract base class, you
 # won't use it for a drawing directly.
 
 # Instead, you should use implementations.
 # See Gdraw2::CairoTrackGroup for detail.
 
 # However, if you want to use drawing backends other than Cairo,
 # you can create your own page class by extending this class,
 # and provide implementation to some methods.
 
 package My::Track;
 
 use Moose;
 use Gdraw2::AbstractTrack;
 
 extends 'Gdraw2::AbstractTrack';
 
 override draw => sub {
     
 };

=head1 DESCRIPTION

Track is a set of elements that are aligned in one line.

=cut

package Gdraw2::AbstractTrack;

use threads::shared;
use Moose;
use Gdraw2::Util::SharedObject;
use Gdraw2::Util::Arranger;
use Gdraw2::Util qw/load_module/;

=head1 EXTENDS

 Gdraw2::Util::SharedObject

=head1 WITH

 Gdraw2::Util::RoleAbstractChild
 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleArrangeable
 Gdraw2::Util::RoleRecordable

=cut

extends 'Gdraw2::Util::SharedObject';
with 'Gdraw2::Util::RoleAbstractChild';
with 'Gdraw2::Util::RoleAbstractParent';
with 'Gdraw2::Util::RoleArrangeable';
with 'Gdraw2::Util::RoleRecordable';

=head1 PUBLIC ATTRIBUTES

=head2 displayName

 desc   : name of this track.
 is     : read/write
 isa    : Str
 default: ""

=cut

has displayName => (
    is        => 'rw',
    metaclass => 'Recordable',
    isa       => 'Str',
    default   => ''
);

=head2 start

 desc   : elements exceed this value will not be shown.
 is     : read/write
 isa    : Num
 default: 1

=cut

has start => (
    is        => 'rw',
    metaclass => 'Recordable',
    isa       => 'Int',
    default   => 1
);

=head2 end

 desc   : elements exceed this value will not be shown.
 is     : read/write
 isa    : Num
 default: 1000

=cut

has end => (
    is        => 'rw',
    metaclass => 'Recordable',
    isa       => 'Int',
    default   => 1000
);

=head2 elementWidth

 desc   : the size of all elements in this track.
 is     : read/write
 isa    : Num
 default: 15

=cut

has elementWidth => (
    is        => 'rw',
    metaclass => 'Recordable',
    isa       => 'Num',
    default   => 15
);

=head2 showGuideLine

 desc   : whether to draw a thick grey line from track start to end.
 is     : read/write
 isa    : Bool
 default: 1

=cut

has 'showGuideLine' => (
    is        => 'rw',
    metaclass => 'Recordable',
    isa       => 'Bool',
    default   => 1
);

=head1 PRIVATE ATTRIBUTES

=head2 trackCascade

=cut

has trackCascade => (
    is  => 'rw',
    isa => 'Num',
);

=head2 arrangers

=cut

has 'arrangers' => (
    is      => 'bare',
    isa     => 'ArrayRef[Gdraw2::Util::Arranger]',
    default => sub { return shared_clone [ Gdraw2::Util::Arranger->new ] }
);

=head2 nameArrangers

=cut

has 'nameArrangers' => (
    is      => 'bare',
    isa     => 'ArrayRef[Gdraw2::Util::Arranger]',
    default => sub { return shared_clone [] }
);

=head1 PUBLIC METHODS

=head2 draw

 usage : $track->draw();
 desc  : render current track. Subclasses that actually perform the drawing
         should override this method.
 args  : none.
 return: none.

=cut

sub draw { confess "need to be overriden" }

sub graphicStart {
    my $self = shift;
    return $self->parent->calcGraphicPosition( $self->start ) -
      $self->parent->bitGraphicLength;
}

sub graphicEnd {
    my $self = shift;
    return $self->parent->calcGraphicPosition( $self->end );
}

sub arrangeOneChild {
    my ( $self, $child ) = @_;
    
    return if !$child->meta->does_role('Gdraw2::Util::RoleArrangeable');

    my $num  = $self->numCascadeLines;
    my $succ = 0;

    my $i = 0;
    foreach my $arranger ( @{ $self->{arrangers} } ) {
        if ( !$arranger->overlaps($child) ) {
            $arranger->addChild($child);
            $child->cascadeLine($i);
            $succ = 1;
            last;
        }
    }
    continue { $i++; }

    if ( !$succ ) {
        my $arranger = Gdraw2::Util::Arranger->new;
        push @{ $self->{arrangers} }, $arranger;
        $arranger->addChild($child);
        $child->cascadeLine($i);
    }

    $child->arranged(1);
}

sub unarrangeOneChild {
    my ( $self, $child ) = @_;
    
    return if !$child->meta->does_role('Gdraw2::Util::RoleArrangeable');

    # remove from arranger
    my $line_num = $self->numCascadeLines;

    for ( my $i = $child->cascadeLine ; $i < $line_num ; $i++ ) {
        my $arranger = $self->{arrangers}[$i];

        # remove from its arranger
        if ( $i == $child->cascadeLine ) {
            $arranger->removeChild($child);
        }

        # remove affected children
        else {
            foreach my $curr ($arranger->getChildren) {
                if ( $curr->overlaps($child) ) {
                    $arranger->removeChild($curr);
                    $curr->arranged(0);
                    $curr->cascadeLine(0);
                }
            }
        }
    }

    $child->arranged(0);
    $child->cascadeLine(0);

    # remove potential lasting empty arrangers
    while (1) {
        if ( $self->{arrangers}[-1]->numChildren == 0 ) {
            if ( @{ $self->{arrangers} } > 1 ) {
                pop @{ $self->{arrangers} };
            }
        }
        else {
            last;
        }
        last if @{ $self->{arrangers} } == 1;
    }
}

sub arrangeOneChildName {
    my ( $self, $child ) = @_;

    if ( $child->showName and length( $child->displayName ) > 0 ) {

        my $succ = 0;
        my $i    = 0;

        foreach my $name_arranger ( @{ $self->{nameArrangers} } ) {
            if ( !$name_arranger->nameOverlaps($child) ) {
                $name_arranger->addChild($child);
                $child->cascadeNameLine($i);
                $succ = 1;
                last;
            }
        }
        continue { $i++ }

        if ( !$succ ) {
            my $name_arranger = Gdraw2::Util::Arranger->new;
            push @{ $self->{nameArrangers} }, $name_arranger;
            $name_arranger->addChild($child);
            $child->cascadeNameLine($i);
        }
    }

    $child->nameArranged(1);
}

sub unarrangeOneChildName {
    my ( $self, $child ) = @_;
    #print "unarrangeOneChildName: amount of arrangers: ",
    #  scalar( @{ $self->{nameArrangers} } ), "\n";

    # traverse through all name arrangers
    my $line_num = $self->numCascadeNameLines;

    for ( my $i = $child->cascadeNameLine ; $i < $line_num ; $i++ ) {
        my $name_arranger = $self->{nameArrangers}[$i];

        # remove from its arranger
        if ( $i == $child->cascadeNameLine ) {
            $name_arranger->removeChild($child);
        }

        # remove affected children
        else {
            foreach my $curr ($name_arranger->getChildren) {
                if ( $curr->nameOverlaps($child) ) {
                    $name_arranger->removeChild($curr);
                    $curr->nameArranged(0);
                    $curr->cascadeNameLine(0);
                }
            }
        }
    }

    $child->nameArranged(0);
    $child->cascadeNameLine(0);

# remove potential lasting arrangers
# notice that the amount of nameline arranger is able to get to 0, unlike the line arranger
    while (1) {
        last if @{ $self->{nameArrangers} } == 0;
        if ( $self->{nameArrangers}[-1]->numChildren == 0 ) {
            pop @{ $self->{nameArrangers} };
        }
        else {
            last;
        }
    }
}

sub arrangeOK {

}

sub calcCascade {
    my $self = shift;
    my $line = shift;
    return $line * $self->elementWidth * 1.2;
}

sub calcNameCascade {
    confess "implement this method in concrete class";
}

sub numCascadeLines {
    return scalar @{ $_[0]->{arrangers} };
}

sub numCascadeLinesInRegin {
    my $self  = shift;
    my $start = shift;
    my $end   = shift;

    my $line = 0;

    foreach my $child ( $self->getChildren ) {
        if (    $child->meta->does_role('Gdraw2::Element::RoleArrangeable')
            and !( $child->start > $end or $start > $child->end )
            and $child->cascadeLine > $line )
        {
            $line = $child->cascadeLine;
        }
    }

    return $line;
}

sub numCascadeNameLines {
    return scalar @{ $_[0]->{nameArrangers} };
}

sub getTrackPage {
    return $_[0]->parent->parent;
}

sub isCircular {
    my $page = $_[0]->getTrackPage;
    confess "track not assigned to a page" if !defined $page;
    return $page->isCircular;
}

sub radius {
    my $self = shift;
    return $self->parent->groupRadius + $self->trackCascade;
}

sub maxContentCascade {
    my $self = shift;
    return $self->calcCascade( $self->numCascadeLines - 1 );
}

sub maxRadius {
    my $self = shift;
    return $self->radius + $self->maxContentCascade + $self->elementWidth / 2;
}

sub arrangeAll {
    my $self = shift;
    if ( $self->isCircular ) {
        foreach my $curr ( $self->getChildren ) {
            $self->arrangeOneChild($curr)
              if $curr->meta->does_role('Gdraw2::Element::RoleArrangeable')
                  and !$curr->arranged;
            $curr->nameArranged(1);
        }
    }
    else {
        foreach my $curr ( $self->getChildren ) {
            $self->arrangeOneChild($curr)
              if $curr->meta->does_role('Gdraw2::Element::RoleArrangeable')
                  and !$curr->arranged;
            $self->arrangeOneChildName($curr) if !$curr->nameArranged;
        }
    }
}

sub unarrangeAll {
    my $self = shift;
    
    $self->{arrangers} = [Gdraw2::Util::Arranger->new];
    $self->{nameArrangers} = [];
    
    foreach my $curr($self->getChildren) {
        $curr->arranged(0)
          if $curr->meta->does_role('Gdraw2::Element::RoleArrangeable');
        $curr->nameArranged(0);
    }
}

around addChildAt => sub {
    my $orig  = shift;
    my $self  = shift;
    my $child = shift;
    my $i     = shift;

    $self->$orig( $child, $i );
    $child->parent($self);

    if ( defined $child->parent and defined $child->parent->parent ) {
        $self->arrangeOneChild($child)
          if $child->meta->does_role('Gdraw2::Element::RoleArrangeable')
              and defined $self->parent
              and defined $self->parent->parent;
        $self->arrangeOneChildName($child)
          if $child->showName
              and $child->displayName ne ''
              and defined $self->parent
              and defined $self->parent->parent
              and !$self->isCircular;
    }
};

around removeChildAt => sub {
    my $orig = shift;
    my $self = shift;
    my $i    = shift;

    my $removed = $self->$orig($i);

    if ( defined $removed ) {

        #$removed->parent(undef);

        if (    $removed->meta->does_role('Gdraw2::Element::RoleArrangeable')
            and $removed->arranged )
        {
            $self->unarrangeOneChild($removed);
        }
        if ( $removed->nameArranged ) {
            $self->unarrangeOneChildName($removed);
        }
    }
    return $removed;
};

around loadHash => sub {
    my $orig  = shift;
    my $self  = shift;
    my $store = shift;

    confess "store don't has children" if !exists $store->{children};
    my @children = @{ $store->{children} };
    delete $store->{children};

    my $obj = $self->$orig( $store, @_ );

    foreach my $child (@children) {
        confess "element store don't has class" if !exists $child->{CLASS};
        my $ele_class = $child->{CLASS};

        load_module($ele_class);
        my $ele = $ele_class->loadHash($child);
        $obj->addChild($ele);
    }

    # restore children
    $store->{children} = [@children];

    return $obj;
};

around toHash => sub {
    my $orig = shift;
    my $self = shift;
    my $hash = $self->$orig(@_);
    $hash->{children} = shared_clone [];

    foreach my $element ( @{ $self->{children} } ) {
        push @{ $hash->{children} }, $element->toHash;
    }

    return $hash;
};

1;

=head1 RECORDABLE ATTRIBUTES

These attributes will be serialized while calling "toHash". See "Gdraw2::Util::RoleRecordable" for details.

 elementWidth
 end
 showGuideLine
 displayName
 start

=head1 SEE ALSO

 Gdraw2::AbstractTrackGroup
 Gdraw2::CairoTrack
 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleAbstractChild
 Gdraw2::Util::RoleRecordable

=cut