=head1 NAME

Gdraw2::AbstractTrackGroup - the base class for a group of tracks that own the same zoom setting.

=head1 SYNOPSIS

 # This is only a abstract base class, you
 # won't use it for a drawing directly.
 
 # Instead, you should use implementations.
 # See Gdraw2::CairoTrackGroup for detail.
 
 # However, if you want to use drawing backends other than Cairo,
 # you can create your own page class by extending this class,
 # and provide implementation to some methods.
 package My::Group;
 
 use Moose;
 use Gdraw2::AbstractTrackGroup;
 
 extends 'Gdraw2::AbstractTrackGroup';
 
 override draw => sub { };

=head1 DESCRIPTION

A group of tracks that own the same zooming. All tracks contained by this group are aligned by "leftMargin" and "rightMargin" by the group's "start" and "end".

=cut

package Gdraw2::AbstractTrackGroup;

use threads::shared;

use Moose;
use Gdraw2::Util::SharedObject;
use Gdraw2::Util::RoleAbstractChild;
use Gdraw2::Util::RoleAbstractParent;
use Gdraw2::Util qw/load_module/;

use Math::Trig qw/pi pi2 pi4 pip2 pip4/;

=head1 EXTENDS

 Gdraw2::Util::SharedObject

=head1 WITH

 Gdraw2::Util::RoleAbstractChild
 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleArrangeable
 Gdraw2::Util::RoleRecordable

=cut

extends 'Gdraw2::Util::SharedObject';
with 'Gdraw2::Util::RoleAbstractChild';
with 'Gdraw2::Util::RoleAbstractParent';
with 'Gdraw2::Util::RoleArrangeable';
with 'Gdraw2::Util::RoleRecordable';

=head1 PUBLIC ATTRIBUTES

=head2 displayName

 desc   : name of this group. Won't be rendered on output, but
          would be showed in the Gdraw2.pl program.
 is     : read/write
 isa    : Str
 default: ""

=cut

has displayName => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => ''
);

=head2 desc

 desc   : useless at current.
 is     : read/write
 isa    : Str
 default: ""

=cut

has desc => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => ''
);

=head2 trackDistance

 desc   : distance between tracks in this group.
 is     : read/write
 isa    : Num
 default: 10

=cut

has trackDistance => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 10
);

=head2 start

 desc   : the sequence position that being aligned to "leftMargin".
 is     : read/write
 isa    : Num
 default: 1

=cut

has start => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Int',
	default   => 1
);

=head2 end

 desc   : the sequence position that being aligned to "rightMargin".
 is     : read/write
 isa    : Num
 default: 1000

=cut

has end => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Int',
	default   => 1000
);

=head2 leftMargin

 desc   : group start will be aligned to this position.
          Only affects linear drawing.
 is     : read/write
 isa    : Num
 default: 50

=cut

has leftMargin => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 50,
);

=head2 rightMargin

 desc   : group end will be aligned to this position.
          Only affects linear drawing.
 is     : read/write
 isa    : Num
 default: 30

=cut

has rightMargin => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 30,
);

=head2 circStartIndent

 desc   : group start will be aligned to this position.
          Only affects circular drawing.
 is     : read/write
 isa    : Num
 default: 0

=cut

has circStartIndent => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 0
);

=head2 circEndIndent

 desc   : group end will be aligned to this position.
          Only affects circular drawing.
 is     : read/write
 isa    : Num
 default: 0

=cut

has circEndIndent => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 0
);

=head1 PRIVATE ATTRIBUTES

=head2 groupCascade

=cut

has groupCascade => (
	is  => 'rw',
	isa => 'Num'
);

=head2 cascadeLine

=cut

has cascadeLine => (
	is      => 'rw',
	isa     => 'Int',
	default => 0
);

=head2 arranged

=cut

has arranged => (
	is      => 'rw',
	isa     => 'Bool',
	default => 0
);

=head1 PUBLIC METHODS

=head2 draw

 usage : $group->draw();
 desc  : render current group. Subclasses that actually perform the drawing
         should override this method.
 args  : none.
 return: none.

=cut

sub draw { confess "need to be overriden" }

=head2 length

 usage : my $length = $group->length();
 desc  : length from start to end.
 args  : none.
 return: group length.

=cut

sub length {
	my $self = shift;

	my $start = $self->start;
	my $end   = $self->end;

	# as nucleotide sequence don't have position 0, we should consider that
	if ( $start * $end > 0 ) {
		return $end - $start + 1;
	}
	else {
		return $end - $start;
	}
}

=head2 trackDisplayWidth

 usage : my $width = $group->trackDisplayWidth();
 desc  : graphical size from left margin to right margin.
 args  : none.
 return: graphical size.

=cut

sub trackDisplayWidth {
	my $self = shift;
	return $self->parent->width - $self->leftMargin - $self->rightMargin;
}

=head2 isCircular

 usage : if ($group->isCircular) {
             # do something
         }
 desc  : if the page being attached on is circular.
 args  : none.
 return: 1 if is circular, 0 if not.

=cut

sub isCircular {
	my $self = shift;
	return $self->parent->isCircular;
}

sub bitGraphicLength {
	my $self = shift;
	if ( $self->isCircular ) {
		return ( pi2- $self->circStartIndent - $self->circEndIndent ) /
		  $self->length;
	}
	else {
		return $self->trackDisplayWidth / $self->length;
	}
}

sub groupRadius {
	my $self = shift;
	return $self->parent->initRadius + $self->groupCascade;
}

sub maxContentCascade {
	my $self       = shift;
	if ($self->numChildren>0) {
		my $last_track = ( $self->getChildren )[-1];
		return $last_track->trackCascade +
		  $last_track->maxContentCascade +
		  $last_track->elementWidth / 2;
	}
	else {
		return 0;
	}
}

sub calcGraphicPosition {
	my ( $self, $pos ) = @_;

	my $start = $self->start;

	if ( $self->isCircular ) {
		if ( $start * $pos > 0 ) {
			return ( $pos - $start + 1 ) *
			  ( pi2 - $self->circStartIndent - $self->circEndIndent ) /
			  $self->length + $self->circStartIndent - pip2;
		}
		else {
			return ( $pos - $start ) *
			  ( pi2 - $self->circStartIndent - $self->circEndIndent ) /
			  $self->length + $self->circStartIndent - pip2;
		}
	}
	else {
		if ( $start * $pos > 0 ) {
			return ( $pos - $start + 1 ) *
			  $self->trackDisplayWidth /
			  $self->length;
		}
		else {
			return ( $pos - $start ) * $self->trackDisplayWidth / $self->length;
		}
	}
}

sub overlaps {
	my $self  = shift;
	my $group = shift;

	if ( $self->isCircular ) {
		return 1
		  if !(
			$self->circStartIndent > pi2- $group->circEndIndent
			or pi2- $self->circEndIndent < $group->circStartIndent
		  );
	}
	else {
		my $width = $self->parent->width;
		return 1
		  if !(
			   $self->leftMargin > $width - $group->rightMargin
			or $width - $self->rightMargin < $group->leftMargin
		  );
	}
	return 0;
}

sub arrangeAll {
	my $self = shift;

	my $page = $self->parent;

	my $i        = 0;
	my $textSize = $self->parent->textSize;

	foreach my $track ( $self->getChildren ) {
		$track->arrangeAll;

		if ( $page->isCircular ) {
			if ( $i == 0 ) {
				$track->trackCascade( $track->elementWidth / 2 );
			}
			else {
				my $prev = $self->{children}[ $i - 1 ];
				$track->trackCascade( $prev->trackCascade +
					  $prev->calcCascade( $prev->numCascadeLines - 1 ) +
					  $prev->elementWidth / 2 +
					  $self->trackDistance +
					  $track->elementWidth / 2 );
			}
		}
		else {
			if ( $i == 0 ) {
				$track->trackCascade(
					$track->calcNameCascade( $track->numCascadeNameLines ) );
			}
			else {
				my $prev = $self->{children}[ $i - 1 ];
				$track->trackCascade( $prev->trackCascade +
					  $prev->calcCascade( $prev->numCascadeLines ) -
					  $prev->elementWidth / 2 +
					  $self->trackDistance +
					  $track->calcNameCascade( $track->numCascadeNameLines ) );
			}
		}
	}
	continue { $i++ }
}

sub unarrangeAll {
    my $self = shift;
    
    foreach my $track ($self->getChildren) {
        $track->unarrangeAll;
    }
}

around addChildAt => sub {
	my ( $orig, $self, $child, $i ) = @_;
	$self->$orig( $child, $i );
	$child->parent($self);
};
around removeChildAt => sub {
	my ( $orig, $self, $i ) = @_;
	my $removed = $self->$orig($i);
	if ( defined $removed ) {
		#$removed->parent(undef);
	}
	return $removed;
};

around applyParams => sub {
    my $orig = shift;
    my $self = shift;
    
    my $page = $self->parent;
    if (defined $page) {
        $page->unarrangeOneGroup($self);
    }
    
    $self->$orig(@_);
};

around loadHash => sub {
	my $orig  = shift;
	my $self  = shift;
	my $store = shift;

	confess "store don't has children" if !exists $store->{children};
	my @children = @{ $store->{children} };
	delete $store->{children};

	my $obj = $self->$orig( $store, @_ );

	foreach my $child (@children) {
		confess "track store don't has class" if !exists $child->{CLASS};
		my $track_class = $child->{CLASS};

		load_module($track_class);
		my $track = $track_class->loadHash($child);
		$obj->addChild($track);
	}
	
	# restore children
	$store->{children} = [@children];

	return $obj;
};

around toHash => sub {
	my $orig = shift;
	my $self = shift;
	my $hash = $self->$orig(@_);
	$hash->{children} = shared_clone [];

	foreach my $track ( @{ $self->{children} } ) {
		push @{ $hash->{children} }, $track->toHash;
	}

	return $hash;
};

1;

=head1 RECORDABLE ATTRIBUTES

These attributes will be serialized while calling "toHash". See "Gdraw2::Util::RoleRecordable" for details.

 trackDistance
 end
 circEndIndent
 circStartIndent
 desc
 rightMargin
 leftMargin
 displayName
 start

=head1 SEE ALSO

 Gdraw2::AbstractTrackPage
 Gdraw2::CairoTrackGroup
 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleAbstractChild
 Gdraw2::Util::RoleRecordable

=cut