=head1 NAME

Gdraw2::AbstractTrackPage - the base class for a page of genomics drawings

=head1 SYNOPSIS

 # This is only a abstract base class, you
 # won't use it for a drawing directly.
 
 # Instead, you should use implementations.
 # See Gdraw2::CairoTrackPage for detail.
 
 # However, if you want to use drawing backends other than Cairo,
 # you can create your own page class by extending this class,
 # and provide implementation to some methods.
 package My::Page;
 
 use Moose;
 use Gdraw2::AbstractTrackPage;
 
 extends 'Gdraw2::AbstractTrackPage';
 
 override draw => sub { };

 override onTitleFontChange => sub { };

 override onTextFontChange => sub { };

 override arrangeCircularElementName => sub { };

=head1 DESCRIPTION

Represents global parameters for a genomics drawing, such as page size, font, etc.

=cut

package Gdraw2::AbstractTrackPage;

use common::sense;
use threads::shared;

use Carp qw/cluck confess/;

use Moose;
use Gdraw2::Util::SharedObject;
use Gdraw2::Util::GroupArranger;
use Gdraw2::Util qw/load_module/;

use List::Util qw/max/;

=head1 EXTENDS

 Gdraw2::Util::SharedObject

=head1 WITH

 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleRecordable

=cut

extends 'Gdraw2::Util::SharedObject';
with 'Gdraw2::Util::RoleAbstractParent';
with 'Gdraw2::Util::RoleRecordable';

=head1 PUBLIC ATTRIBUTES

=head2 groupDistance

 desc   : distance between TrackGroups.
 is     : read/write
 isa    : Num
 default: 10

=cut

has groupDistance => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 10
);

=head2 topMargin

 desc   : distance to the top edge of the page.
 is     : read/write
 isa    : Num
 default: 50

=cut

has topMargin => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 50
);

=head2 width

 desc   : page width.
 is     : read/write
 isa    : Num
 default: 600

=cut

has width => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 600,
	trigger   => sub {
		my $self = shift;
		$self->onPageWidthChange(@_) if $_[0] != $_[1];
	}
);

=head2 height

 desc   : page height.
 is     : read/write
 isa    : Num
 default: 842

=cut

has height => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 842,
	trigger   => sub {
		my $self = shift;
		$self->onPageHeightChange(@_) if $_[0] != $_[1];
	}
);

=head2 isCircular

 desc   : whether to draw the genome in circular.
 is     : read/write
 isa    : Bool
 default: 0 (for false)

=cut

has isCircular => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Bool',
	default   => 0
);

=head2 initRadius

 desc   : when drawing in circular, the radius of innermost track.
 is     : read/write
 isa    : Num
 default: 80

=cut

has initRadius => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 80
);

=head2 titleSize

 desc   : font size for titles.
 is     : read/write
 isa    : Num
 default: 20arrangeCircularElementName

=cut

has titleSize => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 20,
	lazy      => 1,
	trigger   => sub { $_[0]->onTitleFontChange }
);

=head2 titleFont

 desc   : font family for titles.
 is     : read/write
 isa    : Str
 default: Courier for windows, otherwise Courier_new.

=cut

has titleFont => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => $^O eq 'MSWin32' ? 'Courier' : 'Courier_new',
	lazy      => 1,
	trigger   => sub { $_[0]->onTitleFontChange }
);

=head2 textSize

 desc   : font size for texts, such as feature labels.
 is     : read/write
 isa    : Num
 default: 12

=cut

has textSize => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 12,
	lazy      => 1,
	trigger   => sub { $_[0]->onTextFontChange }
);

=head2 textFont

 desc   : font family for texts, such as feature labels.
 is     : read/write
 isa    : Str
 default: Courier for windows, otherwise Courier_new.

=cut

has textFont => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => $^O eq 'MSWin32' ? 'Courier' : 'Courier_new',
	lazy      => 1,
	trigger   => sub { $_[0]->onTextFontChange }
);

=head1 PRIVATE ATTRIBUTES

=head2 groupArrangers

=cut

has groupArrangers => (
	is      => 'bare',
	isa     => 'ArrayRef[Gdraw2::Util::GroupArranger]',
	default => sub { return shared_clone [ Gdraw2::Util::GroupArranger->new ] }
);

around new => sub {
	my $orig  = shift;
	my $class = shift;

	my $self = $class->$orig(@_);

	$self->onTitleFontChange;
	$self->onTextFontChange;
	return $self;
};

around addChildAt => sub {
	my ( $orig, $self, $child, $i ) = @_;
	#say "page add group: $child";
	$self->$orig( $child, $i );
	$child->parent($self);
	$self->arrangeOneGroup($child);
};

around removeChildAt => sub {
	my ( $orig, $self, $i ) = @_;
	my $removed = $self->$orig($i);
	if ( defined $removed ) {

		#$removed->parent(undef);
		if ( $removed->arranged ) {
			$self->unarrangeOneGroup($removed);
		}
	}
	return $removed;
};

around loadHash => sub {
	my $orig  = shift;
	my $self  = shift;
	my $store = shift;

	confess "store don't has children" if !exists $store->{children};
	my @children = @{ $store->{children} };
	delete $store->{children};

	my $obj = $self->$orig( $store, @_ );

	foreach my $child (@children) {
		confess "group store don't has class" if !exists $child->{CLASS};
		my $group_class = $child->{CLASS};

		load_module($group_class);
		my $group = $group_class->loadHash($child);
		$obj->addChild($group);
	}

	return $obj;
};

around toHash => sub {
	my $orig = shift;
	my $self = shift;
	my $hash = $self->$orig(@_);
	$hash->{children} = shared_clone [];

	foreach my $group ( @{ $self->{children} } ) {
		push @{ $hash->{children} }, $group->toHash;
	}

	return $hash;
};

=head1 PUBLIC METHODS

=head2 draw

 usage : $page->draw();
 desc  : draw every thing. Subclasses that actually perform the drawing
         should override this method.
 args  : none.
 return: none.

=cut

sub draw { confess "need to be overriden" }

=head2 getAllElements

 usage : my @elements = $page->getAllElements();
 desc  : recursively access all groups and tracks, and return all
         elements from all tracks.
 args  : none.
 return: an array of elements.

=cut

sub getAllElements {
	my $self = shift;
	my @elements;
	foreach my $grp ( $self->getChildren ) {
		foreach my $trk ( $grp->getChildren ) {
			push @elements, $trk->getChildren;
		}
	}
	return @elements;
}

=head1 PRIVATE METHODS

=cut


sub onPageWidthChange { }

sub onPageHeightChange { }

sub onTitleFontChange {
	confess "This method needs to be overriden";
}

sub onTextFontChange {
	confess "This method needs to be overriden";
}

sub arrangeCircularElementName {
	confess "This method needs to be overrided";
}

sub maxRadius {
	my $self       = shift;
	my $last_grp   = $self->getChildAt(-1);
	my $last_track = $last_grp->getChildAt(-1);
	return $self->initRadius +
	  $last_grp->groupCascade +
	  $last_track->trackCascade +
	  $last_track->calcCascade( $last_track->numCascadeLines );
}

sub arrangeOneGroup {
	my $self  = shift;
	my $group = shift;

	my $num          = $self->numCascadeLines;
	my $arrange_succ = 0;

	my $i = 0;
	foreach my $arranger ( @{ $self->{groupArrangers} } ) {
		if ( !$arranger->overlaps($group) ) {
			$arranger->addChild($group);
			$group->cascadeLine($i);
			$arrange_succ = 1;
			last;
		}
	}
	continue { $i++ }

	if ( !$arrange_succ ) {
		my $arranger = Gdraw2::Util::GroupArranger->new;
		push @{ $self->{groupArrangers} }, $arranger;
		$arranger->addChild($group);

		$group->cascadeLine($i);
	}

	$group->arranged(1);
}

sub unarrangeOneGroup {
	my $self  = shift;
	my $group = shift;
	
	#cluck "unarrange group $group";

	# remove from arranger
	my $line_amount = $self->numCascadeLines;
	for ( my $i = $group->cascadeLine ; $i < $line_amount ; $i++ ) {
		my $arranger = $self->{groupArrangers}[$i];

		# remove from its arranger
		if ( $i == $group->cascadeLine ) {
			$arranger->removeChild($group);
		}

		# remove affected children
		else {
			my $child_num = $arranger->numChildren;

			for ( my $j = 0 ; $j < $child_num ; $j++ ) {
				my $curr = $arranger->getChildAt($j);

				if ( $curr->overlaps($group) ) {
					$arranger->removeChildAt($j);
					$curr->arranged(0);
					$curr->cascadeLine(0);
				}
			}
		}
	}
	$group->arranged(0);
	$group->cascadeLine(0);

	# remove potential lasting empty arrangers
	while (1) {
		if ( $self->{groupArrangers}[-1]->numChildren == 0 ) {
			if ( @{ $self->{groupArrangers} } > 1 ) {
				pop @{ $self->{groupArrangers} };
			}
		}
		else {
			last;
		}
		last if @{ $self->{groupArrangers} } == 1;
	}
}

sub numCascadeLines {
	return scalar @{ $_[0]->{groupArrangers} };
}

sub arrangeAll {
	my $self = shift;

	my @all_groups = $self->getChildren;

	# assign group cascade line
	# arrange tracks before group cascade being calculated
	foreach (@all_groups) {
		$_->arrangeAll;
		$self->arrangeOneGroup($_) if !$_->arranged;
	}

	# calculate group cascade
	my @arrangers = @{ $self->{groupArrangers} };
	for ( my $i = 0 ; $i < @arrangers ; $i++ ) {
		my $arranger = $arrangers[$i];

		# first line
		if ( $i == 0 ) {
			foreach my $group ( $arranger->getChildren ) {
				if ( $self->isCircular ) {
					$group->groupCascade(0);
				}
				else {
					$group->groupCascade( $self->topMargin );
				}
			}
		}

		# not first line
		else {
			my $prev_arranger = $arrangers[ $i - 1 ];
			my $prev_cascade  = $prev_arranger->getChildAt(0)->groupCascade;
			my $prev_max_content_cascade =
			  max( map { $_->maxContentCascade } $prev_arranger->getChildren );
			foreach my $group ( $arranger->getChildren ) {
				$group->groupCascade( $prev_cascade +
					  $prev_max_content_cascade +
					  $self->groupDistance );
			}
		}
	}

	if ( $self->isCircular ) {
		$self->arrangeCircularElementName;
	}
	
}

sub unarrangeAll {
    my $self = shift;
    
    foreach my $arranger ( @{ $self->{groupArrangers} } ) {
        $arranger->removeAllChildren;
    }
    $self->{groupArrangers} = [ Gdraw2::Util::GroupArranger->new ];
    
    foreach my $group ($self->getChildren) {
        $group->arranged(0);
        $group->unarrangeAll;
    }
}
1;

=head1 RECORDABLE ATTRIBUTES

These attributes will be serialized while calling "toHash". See "Gdraw2::Util::RoleRecordable" for details.

 width
 titleSize
 titleFont
 topMargin
 textSize
 height
 isCircular
 textFont
 groupDistance
 initRadius

=head1 SEE ALSO

 Gdraw2::CairoTrackPage
 Gdraw2::Util::RoleAbstractParent
 Gdraw2::Util::RoleRecordable

=cut