=head1 NAME

Gdraw2::CairoTrack - track impelemented using Cairo

=head1 SYNOPSIS
 
 use Gdraw2::CairoTrack;
 
 my $track = Gdraw2::CairoTrack->new(
     start => $track_start,
     end   => $track_end
 );
 
 # add to a group
 group->addChild($track);

=head1 DESCRIPTION

The concrete track class implemented using Cairo 2D drawing library. See Gdraw2::AbstractTrack for more information.

=cut

package Gdraw2::CairoTrack;

use common::sense;
use threads::shared;

use Moose;
use Gdraw2::AbstractTrack;
use Gdraw2::CairoTrackGroup;
use Gdraw2::Util::SharedObject;
use Gdraw2::Util::RoleAbstractChild;
use Gdraw2::Util::RoleAbstractParent;
use Gdraw2::Util::RoleArrangeable;

use Carp qw/cluck confess/;

extends 'Gdraw2::AbstractTrack';

override calcNameCascade => sub {
	my $self = shift;
	my $line = shift;

	my $page   = $self->getTrackPage;
	my $extent = $page->textCairoFont->extents;

	return $self->elementWidth / 2 + $extent->{descent} +
	  $line * ( $extent->{descent} + $extent->{ascent} );
};

override draw => sub {
	my $self = shift;
	my $cr   = $self->getTrackPage->cairoContext;

	# arrange every un-arranged element
	$self->arrangeAll;

	if ( $self->isCircular ) {
		my $angle_start = $self->graphicStart;
		my $angle_end   = $self->graphicEnd;
		if ( $self->showGuideLine ) {
			my $r = $self->radius;
			$cr->save;
			$cr->new_path;
			$cr->set_source_rgba( 0, 0, 0, 0.5 );
			$cr->set_line_width( $self->elementWidth / 5 );
			$cr->arc( 0, 0, $r, $angle_start, $angle_end );
			$cr->stroke;
			$cr->restore;
		}
	}
	else {

		# assume we have been translated to track start position

		# TODO: draw track name

		# draw guide line
		my $gstart = $self->graphicStart;
		my $gend   = $self->graphicEnd;
		if ( $self->showGuideLine ) {
			$cr->save;
			$cr->set_source_rgba( 0, 0, 0, 0.5 );
			$cr->move_to( 0, 0 );
			$cr->rectangle(
				$gstart,
				-$self->elementWidth / 10,
				$gend - $gstart,
				$self->elementWidth / 5
			);
			$cr->fill;
			$cr->restore;
		}
	}

	# draw elements
	my %layers = (
		top     => [],
		bottom  => [],
		default => []
	);

	foreach my $ele ( $self->getChildren ) {
		push @{ $layers{ $ele->layer } }, $ele;
	}

	foreach my $layer (qw/bottom default top/) {
		foreach my $ele ( @{ $layers{$layer} } ) {
			$ele->draw;

			# draw linear names here
			if ( !$self->isCircular ) {
				$ele->drawName;
				$ele->drawNameLeadLine;
			}
		}
	}
};

1;

=head1 SEE ALSO

 Gdraw2::AbstractTrack

=cut