=head1 NAME

Gdraw2::CairoTrackPage - a page of genomics drawings that impelemented using Cairo

=head1 SYNOPSIS
 
 use Gdraw2::CairoTrackPage;
 
 # firstly create a Cairo context
 my $surface = Cairo::PdfSurface->create($file,$width,$height);
 my $context = Cairo::Context->create($surface);
 
 # create the page object using the Cairo context
 my $page = Gdraw2::CairoTrackPage->new(
    width        => $width,
    height       => $height,
    cairoContext => $context
 );
 
 #
 # add groups, tracks
 # create elements
 #
 
 # render everything
 $page->draw;

=head1 DESCRIPTION

The concrete page class implemented using Cairo 2D drawing library. See Gdraw2::AbstractTrackPage for more information.

=cut

package Gdraw2::CairoTrackPage;

use common::sense;
use threads::shared;

use Glib qw/TRUE FALSE/;
use Cairo;

use Moose;
use Gdraw2::AbstractTrackPage;
use Gdraw2::CairoTrackGroup;
use Gdraw2::Util qw/polar_point/;

=head1 EXTENDS

 Gdraw2::AbstractTrackPage

=cut

extends 'Gdraw2::AbstractTrackPage';

=head1 PUBLIC ATTRIBUTES

=head2 cairoContext

 desc   : the Cairo context. See Cairo's documents for more information.
 is     : read/write
 isa    : Cairo::Context
 default: must be given when construction.

=cut

has cairoContext => (
	is       => 'rw',
	isa      => 'Cairo::Context',
	required => 1
);

=head2 titleCairoFont

 desc   : the Cairo font object for titles. See Cairo's documents for
          more information.
 is     : read only
 isa    : Cairo::ScaledFont

=cut

has titleCairoFont => (
	is  => 'ro',
	isa => 'Cairo::ScaledFont',
);

=head2 textCairoFont

 desc   : the Cairo font object for text. See Cairo's documents for
          more information.
 is     : read only
 isa    : Cairo::ScaledFont

=cut

has textCairoFont => (
	is  => 'ro',
	isa => 'Cairo::ScaledFont'
);

# the size of the surface should be changed as width or height changes
my $rebuild_surface = sub {
	my $self = shift;
	say "old surface type:".$self->cairoContext->get_target->get_type;
	my $surface_new = $self->cairoContext->get_target->create_similar('color-alpha',$self->width,$self->height);
	say "new surface type:".$surface_new->get_type;
	my $cr_new = Cairo::Context->create($surface_new);
	$self->cairoContext($cr_new);
};

override onTitleFontChange => sub {
	my $self = shift;
	$self->cairoContext->select_font_face( $self->titleFont, 'normal',
		'normal' );
	$self->cairoContext->set_font_size( $self->titleSize );
	$self->{titleCairoFont} = $self->cairoContext->get_scaled_font;
};

override onTextFontChange => sub {
	my $self = shift;
	$self->cairoContext->select_font_face( $self->textFont, 'normal',
		'normal' );
	$self->cairoContext->set_font_size( $self->textSize );
	$self->{textCairoFont} = $self->cairoContext->get_scaled_font;
};

my $ele_anchor_y = sub {
	my $ele = shift;
	return sin( $ele->anchorGraphicPosition ) * $ele->anchorRadius;
};

my $arrange_one_side = sub {
	my $self = shift;
	my @ele_list = @_;

	my $cf         = $self->textCairoFont;
	my $cf_ext     = $cf->extents;
	my $height     = $cf_ext->{descent} + $cf_ext->{ascent};
	my $max_radius = $self->maxRadius;

	my $max_upper;
	my $max_lower;
	my $upper_i;
	my $lower_i;

	if ( @ele_list % 2 == 0 ) {
		$upper_i = @ele_list / 2 - 1;
		$lower_i = $upper_i + 1;
	}
	else {
		$upper_i = ( @ele_list - 1 ) / 2;
		$lower_i = $upper_i;
	}

	# arrange upper half
	for ( my $i = $upper_i ; $i >= 0 ; $i-- ) {
		my $curr   = $ele_list[$i];
		say "upper element: ",$curr->displayName;
		my $curr_y = &$ele_anchor_y($curr);

		if ( !defined $max_upper ) {
			$max_upper = $curr_y;
		}
		else {
			if ( $curr_y > $max_upper - $height ) {
				$max_upper -= $height;
			}
			else {
				$max_upper = $curr_y;
			}
		}
		$curr->circularNameCascade($max_upper);
	}

	# arrange lower half
	for ( my $i = $lower_i ; $i < @ele_list ; $i++ ) {
		my $curr   = $ele_list[$i];
		say "lower element: ",$curr->displayName;
		my $curr_y = &$ele_anchor_y($curr);

		if ( !defined $max_lower ) {
			$max_lower = $curr_y;
		}
		else {
			if ( $curr_y < $max_lower + $height ) {
				$max_lower += $height;
			}
			else {
				$max_lower = $curr_y;
			}
		}
		$curr->circularNameCascade($max_lower);
	}
};

override arrangeCircularElementName => sub {
	my $self = shift;

	my $last_grp   = $self->getChildAt(-1);
	my $last_track = $last_grp->getChildAt(-1);

	my @all_ele = $self->getAllElements;

	# sort from top to bottom
	@all_ele = sort { &$ele_anchor_y($a) <=> &$ele_anchor_y($b) } @all_ele;

	# divide into right and left side
	my @right_ele;
	my @left_ele;
	foreach my $ele (@all_ele) {
		if ( cos( $ele->anchorGraphicPosition ) > 0 ) { push @right_ele, $ele }
		else                                          { push @left_ele, $ele }
	}

	$self->$arrange_one_side(@right_ele) if @right_ele > 0;
	$self->$arrange_one_side(@left_ele)  if @left_ele > 0;
};

override draw => sub {
	my $self = shift;
	my $cr   = $self->cairoContext;
	
	$cr->new_path;
	$cr->rectangle(0,0,$self->width,$self->height);
	$cr->set_source_rgba(1,1,1,1);
	$cr->fill;

	$self->arrangeAll;

	if ( $self->isCircular ) {
		$cr->save;
		$cr->translate( $self->width / 2, $self->height / 2 );

		foreach my $group ( $self->getChildren ) {
			$group->draw;
		}

		# draw all elements' name
		foreach my $ele ( $self->getAllElements ) {
			$ele->drawName;
			$ele->drawNameLeadLine;
		}

		$cr->restore;
	}
	else {

		# render all groups
		my $group_count = 0;
		foreach my $group ( $self->getChildren ) {
			$cr->save;
			$cr->translate( $group->leftMargin, $group->groupCascade );
			$group->draw;

			$cr->restore;
		}
		continue { $group_count++ }
	}
};

1;

=head1 SEE ALSO

 Gdraw2::AbstractTrackPage

=cut