

package Gdraw2::Element::AbstractElement;

use threads::shared;
use feature qw/say switch/;

use Moose;
use Moose::Util::TypeConstraints;

use Carp qw/cluck/;

use Gdraw2::Util::RGBAColor;
use Gdraw2::Util::SharedObject;
use List::Util qw/max min/;

extends 'Gdraw2::Util::SharedObject';
with 'Gdraw2::Util::RoleAbstractChild';
with 'Gdraw2::Util::RoleArrangeable';
with 'Gdraw2::Util::RoleRecordable';

sub BUILD {
	my $self = shift;
	@{ $self->{position} } = sort { $a <=> $b } @{ $self->{position} };
}

has displayName => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => '',
);

has showName => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Bool',
	default   => 1
);

enum( 'Gdraw2::HorizontalAnchorEnum', qw/left center right/ );

has nameAnchor => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::HorizontalAnchorEnum',
	default   => 'left',
);

has elementAnchor => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::HorizontalAnchorEnum',
	default   => 'left',
);

enum( 'Gdraw2::SimpleLayerEnum', qw/top default bottom/ );

has layer => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::SimpleLayerEnum',
	default   => 'default'
);

has desc => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Str',
	default   => ''
);

has position => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'ArrayRef[Int]',
	default   => sub { shared_clone [1,100] },
	trigger   => sub {
		my ( $self, $list ) = @_;
		@$list = sort { $a <=> $b } @$list;
	}
);

has strand => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Int',
	default   => 0
);

has fillColor => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::Util::RGBColor',
	default   => sub { return Gdraw2::Util::RGBAColor->new( 0, 0.5, 0.5, 1 ) }
);

has strokeColor => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::Util::RGBColor',
	default   => sub { return Gdraw2::Util::RGBAColor->new( 0, 0, 0, 1 ) }
);

has textColor => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::Util::RGBColor',
	default   => sub { return Gdraw2::Util::RGBAColor->new( 0, 0, 0, 1 ) }
);

has lineWidth => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Num',
	default   => 0.5
);

has cascadeNameLine => (
	is      => 'rw',
	isa     => 'Int',
	default => 0
);

has circularNameCascade => (
	is      => 'rw',
	isa     => 'Num',
	default => 0
);

has nameArranged => (
	is      => 'rw',
	isa     => 'Bool',
	default => 0
);

around applyParams => sub {
    my $orig = shift;
    my $self = shift;
    
    my $track = $self->parent;
    if (defined $track) {
        $track->unarrangeOneChild($self);
        $track->unarrangeOneChildName($self);
    }
    
    $self->$orig(@_);
};

sub nameGraphicLength { confess "need to be overriden" }

sub nameGraphicPosition { confess "need to be overriden" }

sub draw { confess "need to be overriden" }

sub drawName { confess "need to be overriden" }

sub drawNameLeadLine { confess "need to be overriden" }

sub start {
	my $self = shift;
	return $self->{position}[0];
}

sub end {
	my $self = shift;
	return $self->{position}[-1];
}

sub positionString {
	my $self = shift;
	return join ', ',@{$self->{position}};
}

sub graphicStart {
	my $self  = shift;
	my $group = $self->getTrackGroup;
	return $group->calcGraphicPosition( $self->start ) -
	  $group->bitGraphicLength;
}

sub graphicEnd {
	my $self = shift;
	return $self->getTrackGroup->calcGraphicPosition( $self->end );
}

sub anchorGraphicPosition {
	my $self = shift;
	given ( $self->elementAnchor ) {
		when ('left') {
			return $self->graphicStart;
		}
		when ('center') {
			return ( $self->graphicStart + $self->graphicEnd ) / 2;
		}
		when ('right') {
			return $self->graphicEnd;
		}
	}
}

sub nameCascade {
	my $self = shift;
	return $self->parent->calcNameCascade( $self->cascadeNameLine );
}

sub nameOverlaps {
	my ( $self, $ele ) = @_;

	my ( $start,     $end )     = $self->nameGraphicPosition;
	my ( $ele_start, $ele_end ) = $ele->nameGraphicPosition;
	if ( $self->isCircular ) {
		die "circular graphic not implemented";
	}
	else {
		if (   $start > $ele_end
			or $ele_start > $end )
		{
			return 0;
		}
		else { return 1 }
	}
}

sub getTrackGroup {
	return $_[0]->parent->parent;
}

sub radius {
	my $self  = shift;
	my $track = $self->parent;
	if ( $self->meta->does_role('Gdraw2::Element::RoleArrangeable') ) {
		return $track->radius + $track->calcCascade( $self->cascadeLine );
	}
	else {
		return $track->radius;
	}
}

sub anchorCascade {
	my $self = shift;
	return $self->parent->elementWidth/2;
}

sub anchorRadius {
	my $self = shift;
	return $self->radius + $self->anchorCascade;
}

sub getTrackPage {
	return $_[0]->parent->parent->parent;
}

sub isCircular {
	return $_[0]->getTrackPage->isCircular;
}

sub arrangeOK {
	my $self = shift;
	if ( $self->meta->does_role('Gdraw2::Element::RoleArrangeable') ) {
		return $self->arranged and $self->nameArranged;
	}
	else {
		return $self->nameArranged;
	}
}

1;
