package Gdraw2::Element::CairoBlock;

use common::sense;
use Moose;
extends 'Gdraw2::Element::AbstractBlock';
with 'Gdraw2::Element::RoleCairoElement';
with 'Gdraw2::Element::RoleCairoDefaultNameLead';

use Gdraw2::Util qw/polar_point/;

override draw => sub {
	my $self = shift;

	my $track = $self->parent;
	my $group = $self->getTrackGroup;

	my $track_start  = $track->start;
	my $track_end    = $track->end;
	my $track_gstart = $track->graphicStart;
	my $track_gend   = $track->graphicEnd;

	# skip if whole block is outside track
	return if $self->start > $track_end or $self->end < $track_start;

	my $cr = $self->getTrackPage->cairoContext;
	$cr->new_path;
	my $w         = $self->parent->elementWidth;
	my $arrow_len = $w / 4;
	my $cascade   = $self->cascade;

	my $numPosition = @{ $self->{position} };

	my $is_circ = $self->isCircular;

	if ( $self->isCircular ) {
		my $r = $self->radius;

		$cr->save;
		$cr->set_line_width( $self->lineWidth );
		$cr->set_line_join('round');

		# traverse through segments
		for ( my $i = 0 ; $i <= $numPosition - 2 ; $i += 2 ) {
			my $seg_start = $self->{position}[$i];
			my $seg_end   = $self->{position}[ $i + 1 ];

			# skip if segment location is outside track
			next if $seg_start > $track_end or $seg_end < $track_start;

			# create path for current segment
			my $gstart =
			  $group->calcGraphicPosition($seg_start) -
			  $group->bitGraphicLength;
			my $gend = $group->calcGraphicPosition($seg_end);

			# short directed blocks
			if ( $self->strand != 0 and $gend - $gstart < $arrow_len / $r ) {
				if ( $i == 0 and $self->strand == -1 ) {
					$cr->move_to( polar_point( $gstart, $r ) );
					$cr->line_to( polar_point( $gend,   $r + $w / 2 ) );
					$cr->line_to( polar_point( $gend,   $r - $w / 2 ) );
				}
				elsif ( $i == $numPosition - 2 and $self->strand == 1 ) {
					$cr->move_to( polar_point( $gstart, $r - $w / 2 ) );
					$cr->line_to( polar_point( $gstart, $r + $w / 2 ) );
					$cr->line_to( polar_point( $gend,   $r ) );
				}
			}

			# long blocks
			else {

				#--- draw left path ---#
				my $angle_start;

				# truncated left path
				if ( $seg_start < $track_start ) {
					$angle_start = $track_gstart;
					$cr->move_to( polar_point( $track_gstart, $r - $w / 2 ) );
					for ( 0.1, 0.3, 0.5, 0.7, 0.9 ) {
						$cr->line_to(
							polar_point(
								$track_gstart - $w / 10 / $r,
								$r + ( $_ - 0.5 ) * $w
							)
						);
						$cr->line_to(
							polar_point(
								$track_gstart, $r + ( $_ - 0.4 ) * $w
							)
						);
					}
				}

				# directed left path
				elsif ( $i == 0 and $self->strand == -1 ) {
					$angle_start = $gstart + $arrow_len / $r;
					$cr->move_to( polar_point( $gstart, $r ) );
				}

				# normal left path
				else {
					$angle_start = $gstart;
				}

				#--- draw right side path ---#
				my $angle_end;

				# truncated right path
				if ( $seg_end > $track_end ) {
					$angle_end = $track_gend;
					$cr->arc( 0, 0, $r + $w / 2, $angle_start, $angle_end );
					for ( 0.1, 0.3, 0.5, 0.7, 0.9 ) {
						$cr->line_to(
							polar_point(
								$angle_end + $w / 10 / $r,
								$r + ( 0.5 - $_ ) * $w
							)
						);
						$cr->line_to(
							polar_point( $angle_end, $r + ( 0.4 - $_ ) * $w ) );
					}
				}

				# directed right path
				elsif ( $i == $numPosition - 2 and $self->strand == 1 ) {
					$angle_end = $gend - $arrow_len / $r;
					$cr->arc( 0, 0,, $r + $w / 2, $angle_start, $angle_end );
					$cr->line_to( polar_point( $gend, $r ) );
				}

				# normal right path
				else {
					$angle_end = $gend;
					$cr->arc( 0, 0,, $r + $w / 2, $angle_start, $angle_end );
				}

				#--- arc back ---#
				$cr->arc_negative( 0, 0, $r - $w / 2, $angle_end,
					$angle_start );

			}
			$cr->close_path;

			# stroke and fill current segment
			$self->fillColor->applyToCairoContext($cr);
			$cr->fill_preserve;
			
			$self->strokeColor->applyToCairoContext($cr);
			$cr->stroke;

			# draw the linking line between multiple segments
			if ( $i != 0 ) {
				my $prev_gend =
				  $group->calcGraphicPosition( $self->{position}[ $i - 1 ] );

				my $mid = ( $prev_gend + $gstart ) / 2;

				# at lower half
				if ( sin($mid) > 0 ) {
					$cr->move_to( polar_point( $prev_gend, $r - $w / 6 ) );
					$cr->line_to( polar_point( $mid,       $r - $w / 2 ) );
					$cr->line_to( polar_point( $gstart,    $r - $w / 6 ) );
				}

				# at upper half
				else {
					$cr->move_to( polar_point( $prev_gend, $r + $w / 6 ) );
					$cr->line_to( polar_point( $mid,       $r + $w / 2 ) );
					$cr->line_to( polar_point( $gstart,    $r + $w / 6 ) );
				}
				
				$self->strokeColor->applyToCairoContext($cr);
				$cr->stroke;
			}
		}
		$cr->restore;
	}
	else {

		$cr->save;
		$cr->set_line_width( $self->lineWidth );
		$cr->set_line_join('round');

		# traverse through segments
		for ( my $i = 0 ; $i <= $numPosition - 2 ; $i += 2 ) {
			my $seg_start = $self->{position}[$i];
			my $seg_end   = $self->{position}[ $i + 1 ];

			# skip if segment location is outside track
			next if $seg_start > $track_end or $seg_end < $track_start;

			my $gstart =
			  $group->calcGraphicPosition($seg_start) - $group->bitGraphicLength;
			my $gend = $group->calcGraphicPosition($seg_end);

			# very short directed segment
			if ( $self->strand != 0 and $gend - $gstart < $arrow_len ) {
				if ( $self->strand == 1 ) {
					$cr->move_to( $gstart, $cascade - $w / 2 );
					$cr->line_to( $gstart, $cascade + $w / 2 );
					$cr->line_to( $gend,   $cascade );
				}
				elsif ( $self->strand == -1 ) {
					$cr->move_to( $gstart, $cascade );
					$cr->line_to( $gend, $cascade + $w / 2 );
					$cr->line_to( $gend, $cascade - $w / 2 );
				}
			}

			# long segment
			else {

				#--- draw left path ---#
				#--- from up to down
				# truncated left path
				if ( $seg_start < $track_start ) {
					$cr->move_to( $track_gstart, $cascade - $w / 2 );
					for ( 0.1, 0.3, 0.5, 0.7, 0.9 ) {
						$cr->line_to( $track_gstart - $w / 10,
							$cascade + ( $_ - 0.5 ) * $w );
						$cr->line_to( $track_gstart, $cascade + ( $_ - 0.4 ) * $w );
					}
				}

				# directed left path
				elsif ( $i == 0 and $self->strand == -1 ) {
					$cr->move_to( $gstart + $arrow_len, $cascade - $w / 2 );
					$cr->line_to( $gstart,              $cascade );
					$cr->line_to( $gstart + $arrow_len, $cascade + $w / 2 );
				}

				# normal left path
				else {
					$cr->move_to($gstart,$cascade-$w/2);
					$cr->line_to($gstart,$cascade+$w/2);
				}

				#--- draw right side path ---#
				#--- from down to up
				# truncated right path
				if ( $seg_end > $track_end ) {
					$cr->line_to($track_gend,$cascade+$w/2);
					for ( 0.1, 0.3, 0.5, 0.7, 0.9 ) {
						$cr->line_to( $track_gend + $w / 10,
							$cascade + ( 0.5-$_ ) * $w );
						$cr->line_to( $track_gend, $cascade + ( 0.4-$_ ) * $w );
					}
				}

				# directed right path
				elsif ( $i == $numPosition - 2 and $self->strand == 1 ) {
					$cr->line_to($gend-$arrow_len,$cascade+$w/2);
					$cr->line_to($gend,$cascade);
					$cr->line_to($gend-$arrow_len,$cascade-$w/2);
				}

				# normal right path
				else {
					$cr->line_to($gend,$cascade+$w/2);
					$cr->line_to($gend,$cascade-$w/2);
				}
			}
			$cr->close_path;

			# stroke and fill
			$self->setContextColor($cr,$self->fillColor);
			$cr->fill_preserve;
			$self->setContextColor($cr,$self->strokeColor);
			$cr->stroke;

			# draw the linking line between multiple segments
			if ( $i != 0 ) {
				my $prev_gend =
				  $group->calcGraphicPosition( $self->{position}[ $i - 1 ] );
				$cr->move_to( $prev_gend, $cascade - $w / 6 );
				$cr->line_to( ( $gstart + $prev_gend ) / 2, $cascade - $w / 2 );
				$cr->line_to( $gstart, $cascade - $w / 6 );
				$self->setContextColor($cr,$self->strokeColor);
				$cr->stroke;
			}
		}

		$cr->restore;
	}
};

1;
