
=head1 NAME 

Gdraw2::Util - Miscellaneous utility functions

=head1 SYNOPSIS

 use Gdraw2::Util qw/polar_point load_model/;

=head1 DESCRIPTION

Contains several utility functions.

=cut

package Gdraw2::Util;

use common::sense;
use Exporter;

use Carp qw/cluck confess/;
use File::Spec::Functions;
use Math::Trig qw/pi pi2 pi4 pip2 pip4/;
use POSIX qw/ceil floor/;

our @ISA = qw/Exporter/;
our @EXPORT_OK =
  qw/polar_point load_module calc_gc_content calc_gc_skew parse_color_str_8 parse_color_str_16 calc_tick_size/;

=head1 FUNCTIONS

=head2 load_module

 usage : load_module('Foo::Bar');
 desc  : load an module at run time, so you won't need to has the module being
         founded at the very beginning.
 args  : an module name.
 return: nothing.

=cut

sub load_module {

    # this function is derived from Bio::Root::Root
    my $name = shift;

    my $module = "_<$name.pm";
    return 1 if $main::{$module};

    # untaint operation for safe web-based running (modified after a fix
    # a fix by Lincoln) HL
    if ( $name !~ /^([\w:]+)$/ ) {
        confess "$name is an illegal perl package name";
    }
    else {
        $name = $1;
    }

    my $load = "$name.pm";

    $load = catfile( split /::/, $load );
    eval { require $load; };
    if ($@) {
        confess "Failed to load module $name. " . $@;
    }
    return 1;
}

=head2 polar_point

 usage : my ($x,$y) = polar_point( $angle, $radius );
 desc  : calculate points of rectangular coordinates from polar coordinates.
 args  : angle, radius
 return: x, y

=cut

sub polar_point {
    my ( $angle, $radius ) = @_;
    return ( $radius * cos $angle, $radius * sin $angle );
}

=head2 calc_gc_content

 usage : my $frac_gc = calc_gc_content($seq);
 desc  : calculate GC content from a nucleotide sequence.
 args  : the sequence.
 return: fraction of GC content.

=cut

sub calc_gc_content {
    my $seq = shift;

    my $num_gc;

    my $len = length $seq;
    for ( my $i = 0 ; $i < $len ; $i++ ) {
        my $char = substr $seq, $i, 1;
        $num_gc++
          if $char eq 'G'
              or $char eq 'C'
              or $char eq 'g'
              or $char eq 'c';
    }

    return $num_gc / $len;
}

=head2 calc_gc_skew

 usage : my $frac_gc_skew = calc_gc_skew($seq);
 desc  : calculate GC skew from a nucleotide sequence.
 args  : the sequence.
 return: fraction of GC skew.

=cut

sub calc_gc_skew {
    my $seq = shift;

    my $num_g;
    my $num_c;

    my $len = length $seq;
    for ( my $i = 0 ; $i < $len ; $i++ ) {
        my $char = substr $seq, $i, 1;
        $num_g++ if $char eq 'G' or $char eq 'g';
        $num_c++ if $char eq 'C' or $char eq 'c';
    }

    return ( $num_g - $num_c ) / ( $num_g + $num_c );
}

=head2 parse_color_str_8

 usage : my ($r, $g, $b) = parse_color_str_8('#rrggbb');
 desc  : parse an GdkColor string, each channel in 8 bit hex.
 args  : the string representation.
 return: red, green, blue color value, each from 0 to 255.

=cut

sub parse_color_str_8 {
    my $str = shift;
    my @result = map { hex($_) } substr( $str, 1, 2 ), substr( $str, 3, 2 ),
      substr( $str, 5, 2 );
    return @result;
}

=head2 parse_color_str_16

 usage : my ($r, $g, $b) = parse_color_str_16('#rrrrggggbbbb');
 desc  : parse an GdkColor string, each channel in 16 bit hex.
 args  : the string representation.
 return: red, green, blue color value, each from 0 to 65535.

=cut

sub parse_color_str_16 {
    my $str = shift;
    my @result = map { hex($_) } substr( $str, 1, 4 ), substr( $str, 5, 4 ),
      substr( $str, 9, 4 );
    return @result;
}

sub calc_tick_size {
    my $length = shift;

    my $scale = floor( log($length) / log(10) );
    my $tick_size = 10**$scale;
    my $num_step = $length/$tick_size;
    
    $tick_size *= 2 if $num_step > 15;
    return $tick_size;
}

1;
