=head1 NAME

Gdraw2::Util::RGBAColor

=head1 SYNOPSIS

 # pure blue, with no transparency
 my $color1 = Gdraw2::Util::RGBAColor->new( 0, 0, 1, 1 );
 
 # dark red, with half transparency
 my $color2 = Gdraw2::Util::RGBColor->new( 0.5, 0, 0, 0.5 );
 
 # get the values in list
 my ($red,$green,$blue,$alpha) = $color1->flat;

=head1 DESCRIPTION

Represents a 4-channel color, each from 0 to 1, corresponding with the Cairo API.

=cut

package Gdraw2::Util::RGBAColor;

use threads::shared;
use Moose;

use Gdraw2::Util::RGBColor;

=head1 EXTENDS

 Gdraw2::Util::RGBColor

=head1 WITH

 Gdraw2::Util::RoleRecordable

=cut

extends 'Gdraw2::Util::RGBColor';
with 'Gdraw2::Util::RoleRecordable';

=head1 PUBLIC ATTRIBUTES

=head2 alpha

 desc   : transparency value, 0 means fully transparent, 1 means opaque.
 is     : read/write
 isa    : number between [ 0, 1 ]
 default: 0

=cut

has alpha => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::RGBColorValue',
	default   => 0
);

=head1 SPECIFIC CONSTRUCTORS

=head2 new( $red, $green, $blue, $alpha )

Gets four color value directly from arguments, so you don't need to write ( red=>$red, green=>$green, blue=>$blue, alpha=>$alpha )

=cut

around BUILDARGS => sub {
	my ( $orig, $class, $red, $green, $blue, $alpha ) = @_;
	my $self = $class->$orig( $red, $green, $blue );
	$self->{alpha} = $alpha;
	return $self;
};

around loadHash => sub {
	my $orig  = shift;
	my $class = shift;
	my $hash  = shift;
	return Gdraw2::Util::RGBAColor->new( $hash->{red}, $hash->{green},
		$hash->{blue}, $hash->{alpha} );
};

override applyToCairoContext => sub {
    my ($self,$cr) = @_;
    $cr->set_source_rgba($self->red,$self->green,$self->blue,$self->alpha);
};

override flat => sub {
	my $self   = shift;
	my @result = super();
	push @result, $self->alpha;
	return @result;
};

override flat_uint16 => sub {
	my $self   = shift;
	my @result = super();
	push @result, int( 65535 * $self->alpha );
	return @result;
};

1;

=head1 RECORDABLE ATTRIBUTES

These attributes will be serialized while calling "toHash". See "Gdraw2::Util::RoleRecordable" for details.

 alpha

=head1 SEE ALSO

 Cairo function set_source_rgba
 Gdraw2::Util::RGBColor
 Gdraw2::Util::RoleRecordable

=cut