=head1 NAME

Gdraw2::Util::RGBColor

=head1 SYNOPSIS

 # pure blue
 my $color1 = Gdraw2::Util::RGBColor->new( 0, 0, 1 );
 
 # dark cyan
 my $color2 = Gdraw2::Util::RGBColor->new( 0, 0.5, 0.5 );
 
 # get the values in list
 my ($red,$green,$blue) = $color1->flat;

=head1 DESCRIPTION

Represents a 3-channel color, each from 0 to 1, corresponding with the Cairo API.

=cut

package Gdraw2::Util::RGBColor;

use threads::shared;

use Moose;
use Moose::Util::TypeConstraints;

use Gdraw2::Util::SharedObject;

=head1 EXTENDS

 Gdraw2::Util::SharedObject

=head1 WITH

 Gdraw2::Util::RoleRecordable

=cut

extends 'Gdraw2::Util::SharedObject';
with 'Gdraw2::Util::RoleRecordable';

subtype 'Gdraw2::RGBColorValue' => as 'Num' => where { $_ >= 0 and $_ <= 1 } =>
  message { "RGB color value must between 0-1: $_" };

=head1 PUBLIC ATTRIBUTES

=head2 red

 desc   : red value
 is     : read/write
 isa    : number between [ 0, 1 ]
 default: 0

=cut

has red => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::RGBColorValue',
	default   => 0,
);

=head2 green

 desc   : green value
 is     : read/write
 isa    : number between [ 0, 1 ]
 default: 0

=cut

has green => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::RGBColorValue',
	default   => 0
);

=head2 blue

 desc   : blue value
 is     : read/write
 isa    : number between [ 0, 1 ]
 default: 0

=cut

has blue => (
	is        => 'rw',
	metaclass => 'Recordable',
	isa       => 'Gdraw2::RGBColorValue',
	default   => 0
);

=head1 SPECIFIC CONSTRUCTORS

=head2 new( $red, $green, $blue )

Gets three color value directly from arguments, so you don't need to write ( red=>$red, green=>$green, blue=>$blue )

=cut

around BUILDARGS => sub {
	my ( $orig, $class, $red, $green, $blue ) = @_;
	return $class->$orig( red => $red, green => $green, blue => $blue );
};

=head1 PUBLIC METHODS

=head1 applyToCairoContext

 usage : $color->applyToCairoContext($cr);
 desc  : apply color to a Cairo context. See Cairo.
 args  : a Cairo context.
 return: none.

=cut

sub applyToCairoContext {
    my ($self,$cr) = @_;
    $cr->set_source_rgb($self->red,$self->green,$self->blue);
}

=head2 flat

 usage : my ($r,$g,$b) = $color->flat;
 desc  : get the color values in a list.
 args  : none.
 return: red, green, blue values.

=cut

sub flat {
	my $self = shift;
	return ( $self->red, $self->green, $self->blue );
}

=head2 flat_uint8

 usage : my ($r,$g,$b) = $color->flat_uint8;
 desc  : get the color values in a list, each is an integer between
         [0,255].
 args  : none.
 return: red, green, blue values.

=cut

sub flat_uint8 {
    my $self = shift;
	return
	  map { int (255 * $_) }
	  ( $self->red, $self->green, $self->blue );
}

=head2 flat_uint16

 usage : my ($r,$g,$b) = $color->flat_uint16;
 desc  : get the color values in a list, each is an integer between
         [0,65535].
 args  : none.
 return: red, green, blue values.

=cut

sub flat_uint16 {
	my $self = shift;
	return
	  map { int (65535 * $_) }
	  ( $self->red, $self->green, $self->blue );
}

=head2 toGdkString8

 usage : my $color_string = $color->toGdkString8;
 desc  : get the color values in "#rrggbb" format which is compatiable
         with GdkColor. See flat_uint8.
 args  : none.
 return: color string.

=cut

sub toGdkString8 {
    my $self = shift;
    my @int_color = $self->flat_uint8;
    my $re =  sprintf("#%02x%02x%02x",@int_color[0,1,2]);
    return $re;
}

=head2 toGdkString8

 usage : my $color_string = $color->toGdkString16;
 desc  : get the color values in "#rrrrggggbbbb" format which is
         compatiable with GdkColor. See flat_uint16.
 args  : none.
 return: color string.

=cut

sub toGdkString16 {
    my $self = shift;
    my @int_color = $self->flat_uint16;
    my $re =  sprintf("#%04x%04x%04x",@int_color[0,1,2]);
    return $re;
}

around loadHash => sub {
	my $orig  = shift;
	my $class = shift;
	my $hash  = shift;
	return Gdraw2::Util::RGBColor->new( $hash->{red}, $hash->{green},
		$hash->{blue} );
};

1;

=head1 RECORDABLE ATTRIBUTES

These attributes will be serialized while calling "toHash". See "Gdraw2::Util::RoleRecordable" for details.

 red
 green
 blue

=head1 SEE ALSO
 
 Cairo function set_source_rgb
 Gdraw2::Util::RGBAColor
 Gdraw2::Util::RoleRecordable

=cut
