package Gdraw2Gtk::App;

use Moose;
use feature qw/say switch/;

use Cairo;
use File::ShareDir qw/:ALL/;
use File::Basename;
use File::Path;
use File::Spec::Functions qw/:ALL/;

use Gdraw2::CairoTrack;
use Gdraw2::CairoTrackGroup;
use Gdraw2::CairoTrackPage;
use Gdraw2::Element::CairoBlock;
use Gdraw2::Element::CairoTagset;
use Gdraw2::Element::CairoDataCurve;
use Gdraw2::Element::CairoMapping;
use Gdraw2::Element::CairoRuler;

use Gdraw2::Util qw/load_module/;

use Gdraw2Gtk::Task;
use Gdraw2Gtk::RichSeqImportConfig;
use Gdraw2Gtk::BlastImportConfig;
use Gdraw2Gtk::EleTypeWizard;
use Gdraw2Gtk::ElementEditor;
use Gdraw2Gtk::TrackEditor;
use Gdraw2Gtk::GroupEditor;
use Gdraw2Gtk::PageEditor;
use Gdraw2Gtk::Act;

use Glib qw/TRUE FALSE/;
use Gtk2 qw/init/;

use Scalar::Util qw/blessed/;

use YAML qw/Load Dump LoadFile DumpFile/;

use constant {
    CLASS_BLOCK   => 'Gdraw2::Element::CairoBlock',
    CLASS_TAGSET  => 'Gdraw2::Element::CairoTagset',
    CLASS_MAPPING => 'Gdraw2::Element::CairoMapping',
    CLASS_RULER   => 'Gdraw2::Element::CairoRuler',
    CLASS_CURVE   => 'Gdraw2::Element::CairoDataCurve',
};

my $PAGEUI_TASK_KEY = 'gdraw2_task_id';
my $TASK_ID_SEED    = 0;

has tasks => (
    is      => 'bare',
    isa     => 'HashRef[Gdraw2Gtk::Task]',
    default => sub { {} }
);

has clip => (
    is      => 'rw',
    default => undef
);

has uiBuilder => (
    is      => 'rw',
    isa     => 'Gtk2::Builder',
    default => sub { return Gtk2::Builder->new }
);

has elementEditor => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::ElementEditor',
    default => sub { Gdraw2Gtk::ElementEditor->new }
);

has eleTypeWizard => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::EleTypeWizard',
    default => sub { Gdraw2Gtk::EleTypeWizard->new }
);

has trackEditor => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::TrackEditor',
    default => sub { Gdraw2Gtk::TrackEditor->new }
);

has groupEditor => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::GroupEditor',
    default => sub { Gdraw2Gtk::GroupEditor->new }
);

has pageEditor => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::PageEditor',
    default => sub { Gdraw2Gtk::PageEditor->new }
);

has richSeqImporter => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::RichSeqImportConfig',
    default => sub { Gdraw2Gtk::RichSeqImportConfig->new }
);

has blastImporter => (
    is      => 'ro',
    isa     => 'Gdraw2Gtk::BlastImportConfig',
    default => sub { Gdraw2Gtk::BlastImportConfig->new }
);

sub BUILD {
    my $self = shift;

    #
    # load GUI resources
    #
    my $file_ui_mw           = dist_file( 'Gdraw2', 'Gdraw2UI.glade' );
    my $file_ui_group_editor = dist_file( 'Gdraw2', 'GroupEditor.glade' );
    my $file_ui_page_editor  = dist_file( 'Gdraw2', 'PageEditor.glade' );
    my $file_ui_dialogs      = dist_file( 'Gdraw2', 'FileDialogs.glade' );

    my $builder = $self->uiBuilder;
    $builder->add_objects_from_file(
        $file_ui_mw, qw/
          AppMW
          AskSaveDialog

          AddElementAction AddGroupAction AddTrackAction DeleteObjAction EditObjAction
          MoveUpAction MoveDownAction
          NewFileAction OpenFileAction SaveFileAction SaveAsAction CloseFileAction
          ImportSeqAction ImportBlastAction ExportFileAction
          ConfigPageAction
          CopyAction CutAction PasteAction
          UndoAction RedoAction
          MainExitAction
          ForceRefreshAction/
    );
    $builder->add_from_file($file_ui_group_editor);
    $builder->add_from_file($file_ui_page_editor);
    $builder->add_from_file($file_ui_dialogs);

    $builder->connect_signals( undef, $self );

    #
    # bind shortcuts
    #
    my $mw = $builder->get_object('AppMW');

    my $accel = Gtk2::AccelGroup->new();
    $accel->connect(
        ord('o'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('OpenFileAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('s'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('SaveFileAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('s'),
        [ 'control-mask', 'shift-mask' ],
        'visible',
        sub {
            $builder->get_object('SaveAsAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('x'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('CutAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('c'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('CopyAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('v'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('PasteAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('z'),
        'control-mask',
        'visible',
        sub {
            $builder->get_object('UndoAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        ord('z'),
        [ 'control-mask', 'shift-mask' ],
        'visible',
        sub {
            $builder->get_object('RedoAction')->activate;
            return TRUE;
        }
    );
    $accel->connect(
        127,
        'release-mask',
        'visible',
        sub {
            $builder->get_object('DeleteObjAction')->activate;
            return TRUE;
        }
    );

    $mw->add_accel_group($accel);

    # update UI state
    $self->_notebook_modify;
}

sub newTask {
    my $self = shift;

    my $builder_mw = $self->uiBuilder;

    # create cairo context
    my $buff_surface = Cairo::ImageSurface->create( 'argb32', 2000, 2000 );
    my $buff_cr = Cairo::Context->create($buff_surface);

    # create TrackPage object
    my $page = Gdraw2::CairoTrackPage->new( cairoContext => $buff_cr );

    if ( $self->pageEditor->run($page) == 0 ) {
        my $id = $TASK_ID_SEED++;

        my $nb_label = Gtk2::Label->new('Untitled');
        my $task     = Gdraw2Gtk::Task->new(
            app   => $self,
            id    => $id,
            page  => $page,
            label => $nb_label
        );
        my $builder = $task->uiBuilder;
        $builder->connect_signals( undef, $self );

        # add to main window
        my $page_ui = $builder->get_object('PageComp');
        $page_ui->{$PAGEUI_TASK_KEY} = $id;
        my $notebook = $builder_mw->get_object('AppMW_Pages');

        my $page_i = $notebook->append_page( $page_ui, $nb_label );
        $notebook->set_tab_reorderable( $page_ui, TRUE );

        my $tree = $builder->get_object('PageComp_PageTree');

        # set DrawingArea size
        my $area = $builder->get_object('PageComp_PreviewArea');
        $area->set_size_request( $page->width, $page->height );

        # store them in the app
        $self->{tasks}{$id} = $task;

        # connect signals
        $area->signal_connect_swapped( 'expose-event',
            \&Gdraw2Gtk::Task::on_area_expose, $task );

        $builder->get_object('but_dbg')
          ->signal_connect( 'clicked', \&dbg_cb, $task );
        $builder->get_object('but_dbg_clip')
          ->signal_connect( 'clicked', \&dbg_clip_cb, $task );

        $task->saved(0);
        $task->page->draw;

        $notebook->set_current_page( $notebook->get_n_pages - 1 );
    }
}

sub openTask {
    my ( $self, $file ) = @_;

    # create cairo context
    my ($store) = LoadFile($file);
    my $buff_surface = Cairo::ImageSurface->create( 'argb32', 2000, 2000 );
    my $buff_cr = Cairo::Context->create($buff_surface);

    # create page object
    my $page =
      Gdraw2::CairoTrackPage->loadHash( $store, cairoContext => $buff_cr );

    # create task object
    my $id       = $TASK_ID_SEED++;
    my $nb_label = Gtk2::Label->new( basename($file) );
    my $task     = Gdraw2Gtk::Task->new(
        app   => $self,
        id    => $id,
        file  => $file,
        page  => $page,
        label => $nb_label,
        saved => 1
    );
    $self->{tasks}{$id} = $task;

    # add to main window
    my $builder    = $task->uiBuilder;
    my $builder_mw = $self->uiBuilder;
    my $page_ui    = $builder->get_object('PageComp');
    $page_ui->{$PAGEUI_TASK_KEY} = $id;
    my $notebook = $builder_mw->get_object('AppMW_Pages');
    $notebook->append_page( $page_ui, $nb_label );

    # set DrawingArea size
    my $area = $builder->get_object('PageComp_PreviewArea');
    $area->set_size_request( $page->width, $page->height );

    # connect signals
    $builder->connect_signals( undef, $self );
    $area->signal_connect_swapped( 'expose-event',
        \&Gdraw2Gtk::Task::on_area_expose, $task );
    $builder->get_object('but_dbg')
      ->signal_connect( 'clicked', \&dbg_cb, $task );
    $builder->get_object('but_dbg_clip')
      ->signal_connect( 'clicked', \&dbg_clip_cb, $task );

    # redraw
    $task->page->draw;
    $task->on_area_expose( undef, $area );

    # swich to that task
    $notebook->set_current_page( $notebook->get_n_pages - 1 );
}

sub closeTask {
    my ( $self, $task ) = @_;

    # remove from store
    delete $self->{tasks}{ $task->id };

    # remove from UI
    my $notebook = $self->uiBuilder->get_object('AppMW_Pages');
    $notebook->remove_page( $notebook->get_current_page );
}

sub getCurrentTaskUI {
    my $self = shift;
    my $book = $self->uiBuilder->get_object('AppMW_Pages');

    if ( $book->get_n_pages > 0 ) {
        return $book->get_nth_page( $book->get_current_page );
    }
    else {
        return undef;
    }
}

sub getCurrentTask {
    my $self = shift;

    my $page_ui = $self->getCurrentTaskUI;
    return undef if !defined $page_ui;

    my $key = $page_ui->{$PAGEUI_TASK_KEY};

    if ( exists $self->{tasks}{$key} ) {
        return $self->{tasks}{$key};
    }
    else {
        confess "failed to fetch Task object from $page_ui via key '$key'";
    }
}

sub areYouSure {
    my $self   = shift;
    my $dialog = Gtk2::MessageDialog->new( undef, 'GTK_DIALOG_MODAL',
        'GTK_MESSAGE_QUESTION', 'GTK_BUTTONS_YES_NO', @_ );
    my $re = $dialog->run;
    $dialog->destroy;
    return $re;
}

sub askForSave {
    my ( $self, $task ) = @_;

    my $file = $task->file;

    my $builder = $self->uiBuilder;

    my $disp_name = $file eq '' ? 'Untitled' : basename($file);
    my $label = $builder->get_object('AskSaveDialog_Label');
    $label->set_text("Save changes for ($disp_name)?");

    my $dialog = $builder->get_object('AskSaveDialog');
    my $re     = $dialog->run;
    $dialog->hide;
    return $re;
}

sub showWarn {
    my $self   = shift;
    my $dialog = Gtk2::MessageDialog->new( undef, 'GTK_DIALOG_MODAL',
        'GTK_MESSAGE_WARNING', 'GTK_BUTTONS_OK', @_ );
    $dialog->run;
    $dialog->destroy;
}

sub showError {
    my $self = shift;
    my $dialog =
      Gtk2::MessageDialog->new( undef, 'GTK_DIALOG_MODAL', 'GTK_MESSAGE_ERROR',
        'GTK_BUTTONS_OK', @_ );
    $dialog->run;
    $dialog->destroy;
}

sub run {
    my $self    = shift;
    my $builder = $self->uiBuilder;
    $builder->get_object('AppMW')->show;
    Gtk2->main();
}

sub on_MainExitAction_activate {
    my $self = shift;

    my $builder     = $self->uiBuilder;
    my $file_dialog = $builder->get_object('SaveTaskFileDialog');

    my $all_saved = 1;

    foreach my $task ( values %{ $self->{tasks} } ) {
        if ( !$task->saved ) {
            my $re = $self->askForSave($task);

            # yes: 0
            if ( $re == 0 ) {
                if ( $task->file ne '' ) {
                    $task->save_task( $task->file );
                }
                else {

                    # ask for a file
                    my $re_file = $file_dialog->run;
                    $file_dialog->hide;

                    if ( $re_file == 0 ) {
                        $task->save_task( $file_dialog->get_filename );
                    }
                    else {
                        $all_saved = 0;
                        last;
                    }
                }
            }

            # no: 1
            elsif ( $re == 1 ) {

                # do nothing if not save
            }

            # cancel: 2
            elsif ( $re == 2 ) {
                $all_saved = 0;
                last;
            }
        }
    }

    Gtk2::main_quit() if $all_saved;

    return TRUE;
}

sub on_NewFileAction_activate {
    my ( $self, $action ) = @_;
    $self->newTask;
}

sub on_OpenFileAction_activate {
    my ( $self, $action ) = @_;

    # select the file to open
    my $file_dialog = Gtk2::FileChooserDialog->new(
        'Open file',
        $self->uiBuilder->get_object('AppMW'),
        'open',
        'Ok'     => 'ok',
        'Cancel' => 'cancel'
    );
    my $filter = Gtk2::FileFilter->new;
    $filter->set_name('yaml');
    $filter->add_pattern('*.yaml');
    $filter->add_pattern('*.yml');
    $file_dialog->add_filter($filter);

    my $re = $file_dialog->run;
    if ( $re eq 'ok' ) {
        my $file = $file_dialog->get_filename;

        # find if this file is already opened
        if ( !file_name_is_absolute($file) ) {
            $file = rel2abs($file);
        }

        foreach my $exist_task ( values %{ $self->{tasks} } ) {
            my $exist_file = $exist_task->file;
            $exist_file = rel2abs($exist_file)
              if !file_name_is_absolute($exist_file);

            if ( $exist_file eq $file ) {
                $file_dialog->destroy;
                $self->showWarn("File '$file' is already opened");
                return;
            }
        }

        # open
        $self->openTask($file);
    }
    $file_dialog->destroy;
}

sub on_SaveAsAction_activate {
    my ( $self, $action ) = @_;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    my $builder = $self->uiBuilder;
    my $dialog  = $builder->get_object('SaveTaskFileDialog');

    my $re = $dialog->run;
    $dialog->hide;

    if ( $re == 0 ) {
        my $file = $dialog->get_filename;
        $task->save_task($file);
    }

}

sub on_SaveFileAction_activate {
    my ( $self, $action ) = @_;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    my $file = $task->file;

    if ( $file eq '' ) {
        $self->on_SaveAsAction_activate(undef);
    }
    else {
        $task->save_task($file);
    }
}

sub on_ImportSeqAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # create file dialog
    my $chooser = Gtk2::FileChooserDialog->new(
        'Import sequence features',
        $self->uiBuilder->get_object('AppMW'),
        'open',
        'Ok'     => 'ok',
        'Cancel' => 'cancel'
    );

    # run
    my $re   = $chooser->run;
    my $file = $chooser->get_filename;
    $chooser->destroy;

    if ( $re eq 'ok' ) {

        # create input handle
        eval { load_module('Bio::SeqIO') };
        if ($@) {
            $self->showError("Failed to load Bioperl module 'Bio::SeqIO':\n$@");
            return undef;
        }

        my $IN;
        eval { $IN = Bio::SeqIO->new( -file => $file ) };
        if ($@) {
            $self->showError("Bioperl failed to parse file '$file':\n$@");
            return undef;
        }

        my $re2 = $self->richSeqImporter->run;
        if ( $re2 == 0 ) {
            eval { $task->importRichSeq( $self->richSeqImporter, $IN ); };
            if ($@) {
                $self->showError("Failed to import '$file':\n$@");
            }
        }

        $IN->close;
    }
}

sub on_ImportBlastAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # create file dialog
    my $chooser = Gtk2::FileChooserDialog->new(
        'Import blast result',
        $self->uiBuilder->get_object('AppMW'),
        'open',
        'Ok'     => 'ok',
        'Cancel' => 'cancel'
    );

    # run
    my $re   = $chooser->run;
    my $file = $chooser->get_filename;
    $chooser->destroy;

    if ( $re eq 'ok' ) {

        # create input handle
        eval { load_module('Bio::SearchIO') };
        if ($@) {
            $self->showError(
                "Failed to load Bioperl module 'Bio::SearchIO':\n$@");
            return undef;
        }

        my $IN;
        eval { $IN = Bio::SearchIO->new( -file => $file ) };
        if ($@) {
            $self->showError("Bioperl failed to parse file '$file':\n$@");
            return undef;
        }

        my $re2 = $self->blastImporter->run;
        if ( $re2 == 0 ) {
            eval { $task->importBlast( $self->blastImporter, $IN ); };
            if ($@) {
                $self->showError("Failed to import '$file':\n$@");
            }
        }

        $IN->close;
    }
}

sub on_ExportFileAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    my $chooser = Gtk2::FileChooserDialog->new(
        'Export image',
        $self->uiBuilder->get_object('AppMW'),
        'save',
        'Ok'     => 'ok',
        'Cancel' => 'cancel'
    );

    my $filter_png = Gtk2::FileFilter->new;
    $filter_png->set_name('Portable Network Graphics (*.png)');
    $filter_png->add_pattern('*.png');
    $filter_png->add_mime_type('image/png');

    my $filter_pdf = Gtk2::FileFilter->new;
    $filter_pdf->set_name('Portable Document Format (*.pdf)');
    $filter_pdf->add_pattern('*.pdf');
    $filter_pdf->add_mime_type('application/pdf');

    my $filter_svg = Gtk2::FileFilter->new;
    $filter_svg->set_name('Scalable Vector Graphics (*.svg)');
    $filter_svg->add_pattern('*.svg');
    $filter_svg->add_mime_type('image/svg+xml');

    my $filter_ps = Gtk2::FileFilter->new;
    $filter_ps->set_name('PostScript (*.ps)');
    $filter_ps->add_pattern('*.ps');

    $chooser->add_filter($filter_pdf);
    $chooser->add_filter($filter_svg);
    $chooser->add_filter($filter_ps);
    $chooser->add_filter($filter_png);

    my $re = $chooser->run;

    if ( $re eq 'ok' ) {
        my $file   = $chooser->get_filename;
        my $filter = $chooser->get_filter;

        my $format;
        if ( $filter == $filter_png ) {
            $format = 'png';
        }
        elsif ( $filter == $filter_pdf ) {
            $format = 'pdf';
        }
        elsif ( $filter == $filter_svg ) {
            $format = 'svg';
        }
        elsif ( $filter == $filter_ps ) {
            $format = 'ps';
        }
        else {
            confess "invalid export filter: $filter";
        }

        eval { $task->export( $file, $format ) };
        if ($@) {
            $self->showError("Failure on export image:\n$@");
        }
    }
    $chooser->destroy;
}

sub on_CloseFileAction_activate {
    my ( $self, $action ) = @_;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    if ( !$task->saved ) {
        my $re = $self->askForSave($task);

        # yes: 0
        if ( $re == 0 ) {
            $self->on_SaveFileAction_activate(undef);
            $self->closeTask($task) if ( $task->saved );
        }

        # no: 1
        elsif ( $re == 1 ) {
            $self->closeTask($task);
        }

        # cancel: 2, do nothing
    }
    else {
        say "";
        $self->closeTask($task);
    }
}

sub on_AddGroupAction_activate {
    my ( $self, $action ) = @_;

    my $task = $self->getCurrentTask;

    # create group
    my $group = Gdraw2::CairoTrackGroup->new();

    # configure group
    my $params = $group->getParams();
    my $re = $self->groupEditor->run( $params, $task->page->isCircular );

    if ( $re == 0 ) {

        # apply parameters
        $group->applyParams($params);

        # get currently selected index
        # add to parent
        my $builder   = $task->uiBuilder;
        my $treeview  = $builder->get_object('PageComp_PageTree');
        my $sel       = $treeview->get_selection;
        my $last_path = ( $sel->get_selected_rows )[-1];

        my $path;
        if ( defined $last_path ) {
            my @i_list = $last_path->get_indices;
            $path = $task->add_group_at( $group, $i_list[0] + 1 );
        }
        else {
            $path = $task->append_group($group);
        }

        # save act
        my $act = Gdraw2Gtk::Act::Add->new(
            task    => $task,
            object  => $group,
            indices => [ $path->get_indices ],
        );
        $task->add_act($act);

        # update
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_AddTrackAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    if ( !defined $last_path ) {
        $self->showWarn("add track: no parent selected");
        return;
    }

    # create track, and config
    my $track  = Gdraw2::CairoTrack->new();
    my $params = $track->getParams;
    my $re     = $self->trackEditor->run($params);

    # get currently selected index, and
    # add track to parent
    if ( $re == 0 ) {
        $track->applyParams($params);
        my @i_list = $last_path->get_indices;

        my $path;

        # add this track after current track
        if ( @i_list > 1 ) {
            $i_list[1]++;
            $path = $task->add_track_at( $track, @i_list[ 0, 1 ] );
        }

        # append this track at end of current group
        else {
            $path = $task->append_track( $track, $i_list[0] );
        }

        # save act
        my $act = Gdraw2Gtk::Act::Add->new(
            task    => $task,
            object  => $track,
            indices => [ $path->get_indices ],
        );
        $task->add_act($act);

        # mark as changed
        $task->saved(0);

        # update preview
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_AddElementAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn(
            "Select a track to add at, or an element to add after!");
        return;
    }

    my @i_list = $last_path->get_indices;
    if ( @i_list < 2 ) {
        $self->showWarn(
            "Select a track to add at, or an element to add after!");
        return;
    }

    # get the type to create
    # create element
    my $ele_type = $self->eleTypeWizard->run;
    my $ele;
    if ( defined $ele_type ) {
        $ele = $ele_type->new;
    }
    else {
        return;
    }

    # configure element
    # add element to parents
    my $params = $ele->getParams;

    if ( $self->elementEditor->run($params) == 0 ) {
        $ele->applyParams($params);

        # add after an element
        my $path;
        if ( @i_list == 3 ) {
            $i_list[2]++;
            $path = $task->add_element_at( $ele, @i_list[ 0, 1, 2 ] );
        }

        # append on tail of a track
        else {
            $path = $task->append_element( $ele, @i_list[ 0, 1 ] );
        }

        # save act
        my $act = Gdraw2Gtk::Act::Add->new(
            task    => $task,
            object  => $ele,
            indices => [ $path->get_indices ],
        );
        $task->add_act($act);

        # mark as unsaved
        $task->saved(0);

        # update preview
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_ConfigPageAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    my $page       = $task->page;
    my $params_old = $page->getParams;

    # run editor
    my $re = $self->pageEditor->run($page);

    if ( $re == 0 ) {
        $task->saved(0);

        # store act
        my $act = Gdraw2Gtk::Act::Config->new(
            task       => $task,
            indices    => [],
            params_old => $params_old,
            params_new => $page->getParams
        );
        $task->add_act($act);

        # change size of DrawingArea
        $task->uiBuilder->get_object('PageComp_PreviewArea')
          ->set_size_request( $page->width, $page->height );

        # update preview
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_EditObjAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    my $page = $task->page;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn("Select an object to config!");
        return;
    }

    my @i_list = $last_path->get_indices;

    if ( @i_list == 3 ) {
        my $ele =
          $page->getChildAt( $i_list[0] )->getChildAt( $i_list[1] )
          ->getChildAt( $i_list[2] );

        my $params = $ele->getParams;

        if ( $self->elementEditor->run($params) == 0 ) {

            # save act
            my $act = Gdraw2Gtk::Act::Config->new(
                task       => $task,
                indices    => [@i_list],
                params_old => $ele->getParams,
                params_new => $params
            );
            $task->add_act($act);

            # apply changes
            $ele->applyParams($params);

            # update GUI
            $task->page->draw;
            $task->on_area_expose( undef, $task->get_preview_area );
            $task->update_table_at(@i_list);
            $task->saved(0);
        }
    }
    elsif ( @i_list == 2 ) {
        my $track  = $page->getChildAt( $i_list[0] )->getChildAt( $i_list[1] );
        my $params = $track->getParams;

        if ( $self->trackEditor->run($params) == 0 ) {

            # save act
            my $act = Gdraw2Gtk::Act::Config->new(
                task       => $task,
                indices    => [@i_list],
                params_old => $track->getParams,
                params_new => $params
            );
            $task->add_act($act);

            # apply changes
            $track->applyParams($params);

            # update preview
            $task->update_table_at(@i_list);
            $task->saved(0);
            $task->page->draw;
            $task->on_area_expose( undef, $task->get_preview_area );
        }
    }
    elsif ( @i_list == 1 ) {
        my $grp    = $page->getChildAt( $i_list[0] );
        my $params = $grp->getParams;

        if ( $self->groupEditor->run( $params, $page->isCircular ) == 0 ) {

            # save act
            my $act = Gdraw2Gtk::Act::Config->new(
                task       => $task,
                indices    => [@i_list],
                params_old => $grp->getParams,
                params_new => $params
            );
            $task->add_act($act);

            # modify group
            $grp->applyParams($params);

            $task->update_table_at(@i_list);
            $task->saved(0);

            # update preview
            $page->arrangeAll;
            $task->page->draw;
            $task->on_area_expose( undef, $task->get_preview_area );
        }
    }
    else {
        confess "invalid indices: @i_list\n";
    }
}

sub on_DeleteObjAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    if ( !defined $last_path ) {
        $self->showWarn("Select an object to delete!");
        return;
    }

    # ask if sure
    if ( $self->areYouSure("Are you sure to delete selected object?") eq 'no' )
    {
        return;
    }

    # delete element, track, or group
    my @i_list = $last_path->get_indices;

    my $obj;
    if ( @i_list == 3 ) {
        $obj = $task->remove_element_at(@i_list);
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
    elsif ( @i_list == 2 ) {
        $obj = $task->remove_track_at(@i_list);
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
    elsif ( @i_list == 1 ) {
        $obj = $task->remove_group_at(@i_list);
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
    else {
        confess "invalid selection index amount: @i_list";
    }

    my $act = Gdraw2Gtk::Act::Remove->new(
        task    => $task,
        indices => [@i_list],
        object  => $obj
    );
    $task->add_act($act);
}

sub on_MoveUpAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;
    my $page = $task->page;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn("Select an object to move!");
        return;
    }

    my @i_list = $last_path->get_indices;
    return if $i_list[-1] == 0;

    # save an act
    my $act = Gdraw2Gtk::Act::Move->new(
        task      => $task,
        indices   => [@i_list],
        direction => 'up'
    );
    $task->add_act($act);

    # do move
    $task->move_at( @i_list, 'up' );

    $task->saved(0);
    $task->page->draw;
    $task->on_area_expose( undef, $task->get_preview_area );
}

sub on_MoveDownAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;
    my $page = $task->page;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn("Select an object to move!");
        return;
    }

    my @i_list = $last_path->get_indices;
    my $amount;

    if ( @i_list == 3 ) {
        my $grp = $page->getChildAt( $i_list[0] );
        my $trk = $grp->getChildAt( $i_list[1] );
        $amount = $trk->numChildren;
    }
    elsif ( @i_list == 2 ) {
        my $grp = $page->getChildAt( $i_list[0] );
        $amount = $grp->numChildren;
    }
    elsif ( @i_list == 1 ) {
        $amount = $page->numChildren;
    }

    return if $i_list[-1] == $amount - 1;

    # save an act
    my $act = Gdraw2Gtk::Act::Move->new(
        task      => $task,
        indices   => [@i_list],
        direction => 'down'
    );
    $task->add_act($act);

    # do move
    $task->move_at( @i_list, 'down' );

    $task->saved(0);
    $task->page->draw;
    $task->on_area_expose( undef, $task->get_preview_area );
}

sub on_CutAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn("Select an object to cut!");
        return;
    }

    my @i_list = $last_path->get_indices;
    my $page   = $task->page;
    my $obj;
    if ( @i_list == 3 ) {
        $obj = $task->remove_element_at(@i_list);
        $self->clip( $obj->toHash );
    }
    elsif ( @i_list == 2 ) {
        $obj = $task->remove_track_at(@i_list);
        $self->clip( $obj->toHash );
    }
    elsif ( @i_list == 1 ) {
        $obj = $task->remove_group_at(@i_list);
        $self->clip( $obj->toHash );
    }
    else {
        confess "invalid index list: @i_list";
    }

    $task->saved(0);

    # save undo/redo
    my $act = Gdraw2Gtk::Act::Remove->new(
        task    => $task,
        indices => [@i_list],
        object  => $obj
    );
    $task->add_act($act);

    # update preview
    $task->page->draw;
    $task->on_area_expose( undef, $task->get_preview_area );
}

sub on_CopyAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # validate selection
    if ( !defined $last_path ) {
        $self->showWarn("Select an object to copy!");
        return;
    }

    my @i_list = $last_path->get_indices;
    my $page   = $task->page;
    if ( @i_list == 3 ) {
        $self->clip( $page->getChildAt( $i_list[0] )->getChildAt( $i_list[1] )
              ->getChildAt( $i_list[2] )->toHash );
    }
    elsif ( @i_list == 2 ) {
        $self->clip(
            $page->getChildAt( $i_list[0] )->getChildAt( $i_list[1] )->toHash );
    }
    elsif ( @i_list == 1 ) {
        $self->clip( $page->getChildAt( $i_list[0] )->toHash );
    }
    else {
        confess "invalid index list: @i_list";
    }
}

sub on_PasteAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;

    return if !defined $task;
    return if !defined $self->clip;

    # get current selection index
    my $builder   = $task->uiBuilder;
    my $treeview  = $builder->get_object('PageComp_PageTree');
    my $sel       = $treeview->get_selection;
    my $last_path = ( $sel->get_selected_rows )[-1];

    # add clip object to task
    my @i_list = defined $last_path ? $last_path->get_indices : ();
    my $obj = $self->clip->{CLASS}->loadHash( $self->clip );
    my $path;
    if ( $obj->does('Gdraw2::Element::RoleCairoElement') ) {
        if ( @i_list == 3 ) {
            $i_list[2]++;
            $path = $task->add_element_at( $obj, @i_list );
        }
        elsif ( @i_list == 2 ) {
            $path = $task->append_element( $obj, @i_list );
        }
        else {
            $self->showWarn(
                "Select a track to append element,\nor a element to add after!"
            );
            return;
        }
    }
    elsif ( $obj->isa('Gdraw2::CairoTrack') ) {
        if ( @i_list >= 2 ) {
            $i_list[1]++;
            $path = $task->add_track_at( $obj, @i_list[ 0, 1 ] );
        }
        elsif ( @i_list == 1 ) {
            $path = $task->append_track( $obj, $i_list[0] );
        }
        else {
            $self->showWarn(
                "Select a group to append track,\nor a track to add after!");
            return;
        }
    }
    elsif ( $obj->isa('Gdraw2::CairoTrackGroup') ) {
        if ( @i_list >= 1 ) {
            $i_list[0]++;
            $path = $task->add_group_at( $obj, $i_list[0] );
        }
        else {
            $path = $task->append_group($obj);
        }
    }
    else {
        confess "invalid clip object: \n", Dump( $self->clip );
    }

    # save undo/redo
    my $act = Gdraw2Gtk::Act::Add->new(
        task    => $task,
        object  => $obj,
        indices => [ $path->get_indices ]
    );
    $task->add_act($act);

    # update task
    $task->saved(0);
    $task->page->draw;
    $task->on_area_expose( undef, $task->get_preview_area );
}

sub on_UndoAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;
    my $re = $task->undo;

    # update UI
    if ($re) {
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_RedoAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;
    my $re = $task->redo;

    # update task
    if ($re) {
        $task->saved(0);
        $task->page->draw;
        $task->on_area_expose( undef, $task->get_preview_area );
    }
}

sub on_ForceRefreshAction_activate {
    my $self = shift;

    my $task = $self->getCurrentTask;
    return if !defined $task;

    say "redo typesetting";

    $task->page->unarrangeAll;
    $task->page->arrangeAll;
    $task->page->draw;
    $task->on_area_expose( undef, $task->get_preview_area );
}

sub _notebook_modify {
    my $self = shift;

    my $builder = $self->uiBuilder;
    my $pages   = $builder->get_object('AppMW_Pages');
    my $state   = $pages->get_n_pages > 0 ? TRUE : FALSE;

    $builder->get_object('AddElementAction')->set_sensitive($state);
    $builder->get_object('AddGroupAction')->set_sensitive($state);
    $builder->get_object('AddTrackAction')->set_sensitive($state);
    $builder->get_object('DeleteObjAction')->set_sensitive($state);
    $builder->get_object('EditObjAction')->set_sensitive($state);
    $builder->get_object('MoveDownAction')->set_sensitive($state);
    $builder->get_object('MoveUpAction')->set_sensitive($state);

    $builder->get_object('SaveFileAction')->set_sensitive($state);
    $builder->get_object('SaveAsAction')->set_sensitive($state);
    $builder->get_object('ImportSeqAction')->set_sensitive($state);
    $builder->get_object('ImportBlastAction')->set_sensitive($state);
    $builder->get_object('ExportFileAction')->set_sensitive($state);
    $builder->get_object('CloseFileAction')->set_sensitive($state);
    $builder->get_object('ConfigPageAction')->set_sensitive($state);
    $builder->get_object('ForceRefreshAction')->set_sensitive($state);

    $builder->get_object('CutAction')->set_sensitive($state);
    $builder->get_object('CopyAction')->set_sensitive($state);
    $builder->get_object('PasteAction')->set_sensitive($state);
    $builder->get_object('RedoAction')->set_sensitive($state);
    $builder->get_object('UndoAction')->set_sensitive($state);

    return TRUE;
}

sub on_AppMW_Pages_page_added {
    my $self = shift;
    return $self->_notebook_modify;
}

sub on_AppMW_Pages_page_removed {
    my $self = shift;
    return $self->_notebook_modify;
}

sub dbg_cb {
    my ( undef, $task ) = @_;

    my $data = $task->page->toHash;
    print Dump $data;
}

sub dbg_clip_cb {

    #print Dump $clip_obj;
}

1;
