package Gdraw2Gtk::ElementEditor;

use strict;
use feature qw/say switch/;
use Moose;
use File::ShareDir qw/:ALL/;
use Glib qw/TRUE FALSE/;
use Gtk2;

use constant {
    CLASS_BLOCK   => 'Gdraw2::Element::CairoBlock',
    CLASS_TAGSET  => 'Gdraw2::Element::CairoTagset',
    CLASS_MAPPING => 'Gdraw2::Element::CairoMapping',
    CLASS_RULER   => 'Gdraw2::Element::CairoRuler',
    CLASS_CURVE   => 'Gdraw2::Element::CairoDataCurve',
};

has uiBuilder => (
    is      => 'ro',
    isa     => 'Gtk2::Builder',
    default => sub { Gtk2::Builder->new }
);

sub BUILD {
    my $self = shift;

    my $builder = $self->uiBuilder;
    my $file_ui = dist_file( 'Gdraw2', 'EleEditor.glade' );
    $builder->add_from_file($file_ui);

    #
    # connect signals
    #
    my $name_entry        = $builder->get_object('EleEditor_NameEntry');
    my $name_check        = $builder->get_object('EleEditor_NameCheck');
    my $name_check_change = sub {
        $name_entry->set_sensitive( $name_check->get_active ? TRUE : FALSE );
    };
    &$name_check_change();
    $name_check->signal_connect( 'toggled', $name_check_change );
}

sub run {
    my ( $self, $params ) = @_;

    my $builder = $self->uiBuilder;

    my $editor      = $builder->get_object('EleEditor');
    my $content_box = $builder->get_object('EleEditor_ContentBox');

    given ( $params->{CLASS} ) {
        when (CLASS_RULER) {
            $content_box->pack_start( $builder->get_object('RulerComp'),
                TRUE, TRUE, 0 );
        }
        when (CLASS_MAPPING) {
            $content_box->pack_start( $builder->get_object('MappingComp'),
                TRUE, TRUE, 0 );
        }
        when (CLASS_CURVE) {
            $content_box->pack_start( $builder->get_object('CurveComp'),
                TRUE, TRUE, 0 );
        }
    }

    $editor->signal_connect(
        'delete-event',
        sub {
            $editor->response(1);
        }
    );

    # get widgets from builder
    my $name_entry = $builder->get_object('EleEditor_NameEntry');
    my $name_check = $builder->get_object('EleEditor_NameCheck');
    my $desc_entry = $builder->get_object('EleEditor_DescEntry');
    my $pos_entry  = $builder->get_object('EleEditor_PosEntry');

    my $strand_fwd_rbut = $builder->get_object('EleStrand_FwdRadio');
    my $strand_rev_rbut = $builder->get_object('EleStrand_RevRadio');
    my $strand_non_rbut = $builder->get_object('EleStrand_NonRadio');

    my $stroke_color_but = $builder->get_object('EleEditor_StrokeColorBut');
    my $fill_color_but   = $builder->get_object('EleEditor_FillColorBut');
    my $text_color_but   = $builder->get_object('EleEditor_TextColorBut');

    my $ele_anchor_left_rbut   = $builder->get_object('EleAnchor_LeftRadio');
    my $ele_anchor_center_rbut = $builder->get_object('EleAnchor_CenterRadio');
    my $ele_anchor_right_rbut  = $builder->get_object('EleAnchor_RightRadio');

    my $name_anchor_left_rbut = $builder->get_object('NameAnchor_LeftRadio');
    my $name_anchor_center_rbut =
      $builder->get_object('NameAnchor_CenterRadio');
    my $name_anchor_right_rbut = $builder->get_object('NameAnchor_RightRadio');

    my $layer_top_rbut     = $builder->get_object('EleLayer_TopRadio');
    my $layer_default_rbut = $builder->get_object('EleLayer_DefaultRadio');
    my $layer_bottom_rbut  = $builder->get_object('EleLayer_BottomRadio');

    # set element properties to widgets
    $name_entry->set_text( $params->{displayName} );

    if   ( $params->{showName} ) { $name_check->set_active(TRUE) }
    else                         { $name_check->set_active(FALSE) }

    $desc_entry->set_text( $params->{desc} );
    $pos_entry->set_text( join( ', ', @{ $params->{position} } ) );

    given ( $params->{strand} ) {
        when (-1) { $strand_rev_rbut->set_active(TRUE) }
        when (0)  { $strand_non_rbut->set_active(TRUE) }
        when (1)  { $strand_fwd_rbut->set_active(TRUE) }
        default {
            confess 'element <'
              . $params->{displayName}
              . "> has invalid element strand: "
              . $params->{strand};
        }
    }

    $stroke_color_but->set_color(
        Gtk2::Gdk::Color->new(
            map { int 65535 * $_ }
              @{ $params->{strokeColor} }{qw/red green blue/}
        )
    );
    $stroke_color_but->set_alpha( int 65535 * $params->{strokeColor}{alpha} );

    $fill_color_but->set_color(
        Gtk2::Gdk::Color->new(
            map { int 65535 * $_ } @{ $params->{fillColor} }{qw/red green blue/}
        )
    );
    $fill_color_but->set_alpha( int 65535 * $params->{fillColor}{alpha} );

    $text_color_but->set_color(
        Gtk2::Gdk::Color->new(
            map { int 65535 * $_ } @{ $params->{textColor} }{qw/red green blue/}
        )
    );
    $fill_color_but->set_alpha( int 65535 * $params->{textColor}{alpha} );

    given ( $params->{elementAnchor} ) {
        when ('left')   { $ele_anchor_left_rbut->set_active(TRUE) }
        when ('center') { $ele_anchor_center_rbut->set_active(TRUE) }
        when ('right')  { $ele_anchor_right_rbut->set_active(TRUE) }
    }

    given ( $params->{nameAnchor} ) {
        when ('left')   { $name_anchor_left_rbut->set_active(TRUE) }
        when ('center') { $name_anchor_center_rbut->set_active(TRUE) }
        when ('right')  { $name_anchor_right_rbut->set_active(TRUE) }
    }

    given ( $params->{layer} ) {
        when ('top')     { $layer_top_rbut->set_active(TRUE) }
        when ('default') { $layer_default_rbut->set_active(TRUE) }
        when ('bottom')  { $layer_bottom_rbut->set_active(TRUE) }
    }

    # special element, add special dialog components
    if ( $params->{CLASS} eq CLASS_RULER ) {
        $builder->get_object('RulerComp_TickSizeSpin')
          ->set_value( $params->{tickSize} );
        $builder->get_object('RulerComp_GridLengthSpin')
          ->set_value( $params->{gridLength} );
        $builder->get_object('RulerComp_IntTickCheck')
          ->set_active( $params->{integralTick} ? TRUE : FALSE );
        $builder->get_object('RulerComp_AbbrevTickCheck')
          ->set_active( $params->{abbrevTickText} ? TRUE : FALSE );
    }
    if ( $params->{CLASS} eq CLASS_MAPPING ) {
        $builder->get_object('MappingComp_ExtGroupSpin')
          ->set_value( $params->{extGroupIndex} );
        $builder->get_object('MappingComp_ExtTrackSpin')
          ->set_value( $params->{extTrackIndex} );
        $builder->get_object('MappingComp_ExtStartSpin')
          ->set_value( $params->{extStart} );
        $builder->get_object('MappingComp_ExtEndSpin')
          ->set_value( $params->{extEnd} );
    }
    if ( $params->{CLASS} eq CLASS_CURVE ) {
        $builder->get_object('CurveComp_FillColorBut')->set_color(
            Gtk2::Gdk::Color->new(
                map { int 65535 * $_ }
                  @{ $params->{fillColor2} }{qw/red green blue/}
            )
        );
        $builder->get_object('CurveComp_FillColorBut')->set_alpha(
            int 65535 * $params->{fillColor2}{alpha}
        );
        $builder->get_object('CurveComp_DataEntry')
          ->set_text( join ",", @{ $params->{data} } );
        $builder->get_object('CurveComp_HCenterSpin')
          ->set_value( $params->{middleLine} );
    }

    #
    # run dialog
    #
    my $re = $editor->run;

    #
    # apply changes
    #
    say "config element: get response $re";
    if ( $re == 0 ) {
        $params->{displayName} = $name_entry->get_text;
        $params->{showName}    = $name_check->get_active ? 1 : 0;
        $params->{desc}        = $desc_entry->get_text;

        my @pos_nums = split( /[^\-0-9]+/, $pos_entry->get_text );
        if ( @pos_nums < 1 ) {
            warn "invalid position list: (@pos_nums), set to (1,1)";
            @pos_nums = ( 1, 1 );
        }
        @{ $params->{position} } = @pos_nums;

        if    ( $strand_rev_rbut->get_active ) { $params->{strand} = -1 }
        elsif ( $strand_non_rbut->get_active ) { $params->{strand} = 0 }
        elsif ( $strand_fwd_rbut->get_active ) { $params->{strand} = 1 }

        my $color_stroke = $stroke_color_but->get_color;
        $params->{strokeColor}{red}   = $color_stroke->red / 65535;
        $params->{strokeColor}{green} = $color_stroke->green / 65535;
        $params->{strokeColor}{blue}  = $color_stroke->blue / 65535;
        $params->{strokeColor}{alpha} = $stroke_color_but->get_alpha / 65535;

        my $color_fill = $fill_color_but->get_color;
        $params->{fillColor}{red}   = $color_fill->red / 65535;
        $params->{fillColor}{green} = $color_fill->green / 65535;
        $params->{fillColor}{blue}  = $color_fill->blue / 65535;
        $params->{fillColor}{alpha} = $fill_color_but->get_alpha / 65535;
        
        my $color_text = $text_color_but->get_color;
        $params->{textColor}{red}   = $color_text->red / 65535;
        $params->{textColor}{green}   = $color_text->green / 65535;
        $params->{textColor}{blue}   = $color_text->blue / 65535;
        $params->{textColor}{alpha}   = $text_color_but->get_alpha / 65535;

        if ( $ele_anchor_left_rbut->get_active ) {
            $params->{elementAnchor} = 'left';
        }
        elsif ( $ele_anchor_center_rbut->get_active ) {
            $params->{elementAnchor} = 'center';
        }
        elsif ( $ele_anchor_right_rbut->get_active ) {
            $params->{elementAnchor} = 'right';
        }

        if ( $name_anchor_left_rbut->get_active ) {
            $params->{nameAnchor} = 'left';
        }
        elsif ( $name_anchor_center_rbut->get_active ) {
            $params->{nameAnchor} = 'center';
        }
        elsif ( $name_anchor_right_rbut->get_active ) {
            $params->{nameAnchor} = 'right';
        }

        if ( $layer_top_rbut->get_active ) {
            $params->{layer} = 'top';
        }
        elsif ( $layer_default_rbut->get_active ) {
            $params->{layer} = 'default';
        }
        elsif ( $layer_bottom_rbut->get_active ) {
            $params->{layer} = 'bottom';
        }

        if ( $params->{CLASS} eq CLASS_RULER ) {
            $params->{tickSize} =
              $builder->get_object('RulerComp_TickSizeSpin')->get_value;
            $params->{gridLength} =
              $builder->get_object('RulerComp_GridLengthSpin')->get_value;
            $params->{integralTick} =
              $builder->get_object('RulerComp_IntTickCheck')->get_active
              ? 1
              : 0;
            $params->{abbrevTickText} =
              $builder->get_object('RulerComp_AbbrevTickCheck')->get_active
              ? 1
              : 0;
        }
        if ( $params->{CLASS} eq CLASS_MAPPING ) {
            $params->{extGroupIndex} =
              $builder->get_object('MappingComp_ExtGroupSpin')->get_value;
            $params->{extTrackIndex} =
              $builder->get_object('MappingComp_ExtTrackSpin')->get_value;
            $params->{extStart} =
              $builder->get_object('MappingComp_ExtStartSpin')->get_value;
            $params->{extEnd} =
              $builder->get_object('MappingComp_ExtEndSpin')->get_value;
        }
        if ( $params->{CLASS} eq CLASS_CURVE ) {
            @{ $params->{data} } = split /[^\-0-9]+/,
              $builder->get_object('CurveComp_DataEntry')->get_text;
            $params->{middleLine} =
              $builder->get_object('CurveComp_HCenterSpin')->get_value;
            my $color_fill_2 =
              $builder->get_object('CurveComp_FillColorBut')->get_color;
            $params->{fillColor2}{red}   = $color_fill_2->red / 65535;
            $params->{fillColor2}{green} = $color_fill_2->green / 65535;
            $params->{fillColor2}{blue}  = $color_fill_2->blue / 65535;
            $params->{fillColor2}{alpha} =
              $builder->get_object('CurveComp_FillColorBut')->get_alpha / 65535;
        }
    }

    $editor->hide;
    return $re;
}

1;

