package Gdraw2Gtk::Task;

use common::sense;
use Cairo;
use Carp qw/cluck confess/;
use File::Basename;
use File::ShareDir qw/:ALL/;
use Glib qw/TRUE FALSE/;

use Gdraw2::Util qw/calc_gc_content calc_gc_skew calc_tick_size/;

use Gdraw2Gtk::Act;

use Moose;
use YAML::Syck qw/Load Dump LoadFile DumpFile/;

use constant {
    COL_NAME      => 0,
    COL_TYPE      => 1,
    COL_LOCATION  => 2,
    COL_STROKE    => 3,
    COL_FILL      => 4,
    COL_COLOR_SET => 5
};
use constant { FILE_GLADE    => 'Gdraw2UI.glade' };
use constant { MAX_UNDO_REDO => 20 };

has app => (
    is       => 'ro',
    isa      => 'Gdraw2Gtk::App',
    weak_ref => 1,
    required => 1
);

has id => (
    is       => 'rw',
    isa      => 'Str',
    required => 1
);

has file => (
    is      => 'rw',
    isa     => 'Str',
    default => ''
);

has page => (
    is       => 'rw',
    isa      => 'Gdraw2::AbstractTrackPage',
    required => 1
);

my $unsave_prefix = '*';
has saved => (
    is      => 'rw',
    isa     => 'Bool',
    default => 0,
    trigger => sub {
        my $self = shift;

        my $label      = $self->label;
        my $label_text = $label->get_text;

        if ( $self->saved ) {
            $label_text =~ s/^\Q$unsave_prefix//;
        }
        else {
            $label_text = $unsave_prefix . $label_text
              if !( $label_text =~ /^\Q$unsave_prefix/ );
        }

        $label->set_text($label_text);
    }
);

has uiBuilder => (
    is      => 'rw',
    isa     => 'Gtk2::Builder',
    default => sub { return Gtk2::Builder->new }
);

has label => (
    is       => 'rw',
    isa      => 'Object',
    required => 1
);

has undo_list => (
    is      => 'rw',
    isa     => 'ArrayRef[Gdraw2Gtk::Act]',
    default => sub { return [] }
);

has redo_list => (
    is      => 'rw',
    isa     => 'ArrayRef[Gdraw2Gtk::Act]',
    default => sub { return [] }
);

sub BUILD {
    my $self = shift;

    my $builder = $self->uiBuilder;

    # load UI resources
    my $file_ui_mw = dist_file( 'Gdraw2', 'Gdraw2UI.glade' );
    $builder->add_objects_from_file(
        $file_ui_mw, qw/PageComp
          PageTreeModel
          AddObjMenu
          AddGroupAction
          AddTrackAction
          AddElementAction
          DeleteObjAction
          EditObjAction
          MoveUpAction
          MoveDownAction/
    );

    # bind popup menu
    my $menu_add = $builder->get_object('AddObjMenu');
    $builder->get_object('PageComp_AddBut')->set_menu($menu_add);

    # setup treeview
    my $col_name   = $builder->get_object('PageTree_NameCol');
    my $col_loc    = $builder->get_object('PageTree_LocationCol');
    my $col_stroke = $builder->get_object('PageTree_StrokeColorCol');
    my $col_fill   = $builder->get_object('PageTree_FillColorCol');

    my $r_name = Gtk2::CellRendererText->new;
    $col_name->pack_start( $r_name, TRUE );
    $col_name->add_attribute( $r_name, 'text', COL_NAME );

    my $r_loc = Gtk2::CellRendererText->new;
    $col_loc->pack_start( $r_loc, TRUE );
    $col_loc->add_attribute( $r_loc, 'text', COL_LOCATION );

    my $r_stroke = Gtk2::CellRendererText->new;
    $col_stroke->pack_start( $r_stroke, TRUE );
    $col_stroke->add_attribute( $r_stroke, 'cell-background', COL_STROKE );
    $col_stroke->add_attribute( $r_stroke, 'cell-background-set',
        COL_COLOR_SET );

    my $r_fill = Gtk2::CellRendererText->new;
    $col_fill->pack_start( $r_fill, TRUE );
    $col_fill->add_attribute( $r_fill, 'cell-background',     COL_FILL );
    $col_fill->add_attribute( $r_fill, 'cell-background-set', COL_COLOR_SET );

    # update treeview
    $self->rebuild_treeview;
}

sub get_tree_view {
    my $self = shift;
    return $self->uiBuilder->get_object('PageComp_PageTree');
}

sub get_preview_area {
    my $self = shift;
    return $self->uiBuilder->get_object('PageComp_PreviewArea');
}

sub get_tree_model {
    my $self = shift;
    return $self->uiBuilder->get_object('PageTreeModel');
}

sub get_object_at {
    my ( $self, @indices ) = @_;

    if ( @indices == 3 ) {
        return $self->page->getChildAt( $indices[0] )->getChildAt( $indices[1] )
          ->getChildAt( $indices[2] );
    }
    elsif ( @indices == 2 ) {
        return $self->page->getChildAt( $indices[0] )
          ->getChildAt( $indices[1] );
    }
    elsif ( @indices == 1 ) {
        return $self->page->getChildAt( $indices[0] );
    }
    elsif ( @indices == 0 ) {
        return $self->page;
    }
}

sub move_at {
    my ( $self, @args ) = @_;

    my $direction = pop @args;
    my @i_to      = @args;

    if ( $direction eq 'up' ) {
        $i_to[-1]--;
    }
    elsif ( $direction eq 'down' ) {
        $i_to[-1]++;
    }
    else {
        confess "invalid direction: '$direction'";
    }

    # get parent
    my $page = $self->page;

    my $parent;
    if ( @args == 3 ) {
        $parent = $page->getChildAt( $args[0] )->getChildAt( $args[1] );
    }
    elsif ( @args == 2 ) {
        $parent = $page->getChildAt( $args[0] );
    }
    elsif ( @args == 1 ) {
        $parent = $page;
    }
    else {
        confess "invalid arguments: @args";
    }

    # validate object to move
    if ( $direction eq 'up' and $args[-1] == 0 ) {
        confess "attempt to move up first object";
    }
    if ( $direction eq 'down' and $args[-1] == $parent->numChildren - 1 ) {
        confess "attempt to move down last object at $args[-1]";
    }

    # modify the Gdraw2 tree
    my $obj = $parent->removeChildAt( $args[-1] );
    $parent->addChildAt( $obj, $i_to[-1] );

    # modify the table
    my $model    = $self->uiBuilder->get_object('PageTreeModel');
    my $iter_obj = $model->get_iter_from_string( join ':', @args );
    my $iter_to  = $model->get_iter_from_string( join ':', @i_to );
    $model->swap( $iter_obj, $iter_to );

    # move cursor
    my $path  = $model->get_path($iter_obj);
    my $table = $self->uiBuilder->get_object('PageComp_PageTree');
    $table->set_cursor($path);
}

sub save_task {
    my $self = shift;
    my $file = shift;

    # save
    $file .= '.yaml' if !$file =~ /\.yaml$/;
    my $page = $self->page;
    DumpFile( $file, $page->toHash );

    # properties
    $self->file($file);
    $self->label->set_text( basename($file) );
    $self->saved(1);
}

sub importRichSeq {
    my ( $self, $config, $IN ) = @_;

    # generate feature-track mapping
    my @track_names;
    my %feature_map
      ;    # feature => [ track_i, ele_type, [tags], [stroke], [fill] ]

    $config->getTrackTemplates( \@track_names, \%feature_map );

    # traverse input
    my @groups;

    while (1) {

        # get seq
        my $seq_obj;
        eval { $seq_obj = $IN->next_seq };
        if ($@) {
            $self->app->showError($@);
            return;
        }
        last if !$seq_obj;

        my $len = $seq_obj->length;

        # check if we import parital
        my $start;
        my $end;
        if ( $config->importPartial ) {
            ( $start, $end ) =
              sort { $a <=> $b } ( $config->partialStart, $config->partialEnd );
            $end = $len if $len < $end;
        }
        else {
            $start = 1;
            $end   = $len;
        }

        # create group
        my $group = Gdraw2::CairoTrackGroup->new(
            displayName => $seq_obj->desc,
            start       => $start,
            end         => $end
        );
        push @groups, $group;

        # create element tracks
        foreach my $track_name (@track_names) {
            my $track = Gdraw2::CairoTrack->new(
                start       => $start,
                end         => $end,
                displayName => $track_name
            );
            $group->addChild($track);
        }

        # traverse through features
        foreach my $feat ( $seq_obj->get_SeqFeatures ) {
            my $type = $feat->primary_tag;
            next if !exists $feature_map{$type};
            next
              if ( $feat->end < $start )
              or ( $end < $feat->start );

            foreach my $curr ( @{ $feature_map{$type} } ) {
                my ( $track_i, $ele_type, $tags, $stroke, $fill ) = @$curr;
                my $track = $group->getChildAt($track_i);

                # generate element display name
                my @ele_name_values;
                foreach my $tag (@$tags) {
                    next if !$feat->has_tag($tag);
                    push @ele_name_values, $feat->get_tag_values($tag);
                }
                my $ele_name = join ' ', @ele_name_values;

                # generate element
                my $ele;
                if ( $ele_type eq 'block' ) {
                    $ele = Gdraw2::Element::CairoBlock->new(
                        displayName => $ele_name,
                        position    => [ $feat->start, $feat->end ],
                        strand      => $feat->strand,
                        strokeColor => Gdraw2::Util::RGBAColor->new(@$stroke),
                        fillColor   => Gdraw2::Util::RGBAColor->new(@$fill)
                    );
                }
                elsif ( $ele_type eq 'tagset' ) {
                    $ele = Gdraw2::Element::CairoBlock->new(
                        displayName => $ele_name,
                        position =>
                          [ int( ( $feat->start + $feat->end ) / 2 ) ],
                        strokeColor => Gdraw2::Util::RGBAColor->new(@$stroke),
                        fillColor   => Gdraw2::Util::RGBAColor->new(@$fill)
                    );
                }
                else {
                    confess "invalid element type: '$ele_type'";
                }

                # append element to track
                $track->addChild($ele);
            }
        }

        # draw ruler
        if ( $config->drawRuler ) {
            my $track_ruler = Gdraw2::CairoTrack->new(
                start         => $start,
                end           => $end,
                showGuideLine => 0,
                displayName   => 'ruler track'
            );
            $group->addChildAt( $track_ruler, 0 );

            my $ruler = Gdraw2::Element::CairoRuler->new(
                displayName => 'ruler',
                position    => [ $start, $end ],
                gridLength  => $config->rulerTickInterval,
            );
            $track_ruler->addChild($ruler);
        }

        # draw GC content
        if ( $config->drawGC ) {
            my $track_gc = Gdraw2::CairoTrack->new(
                displayName   => 'GC content track',
                start         => $start,
                end           => $end,
                showGuideLine => 0
            );
            $group->addChild($track_gc);

            my @data;
            my $gc_start;
            my $gc_end;
            for (
                $gc_start = $start ;
                $gc_start < $end ;
                $gc_start += $config->gcWindowStep
              )
            {
                $gc_end = $gc_start + $config->gcWindowSize - 1;
                last if $gc_end > $end;
                push @data,
                  calc_gc_content( $seq_obj->subseq( $gc_start, $gc_end ) );
            }

            $track_gc->addChild(
                Gdraw2::Element::CairoDataCurve->new(
                    displayName => 'GC content',
                    position    => [ $start, $gc_end ],
                    middleLine  => 0.5,
                    data        => [@data]
                )
            );

        }

        # draw GC skew
        if ( $config->drawSkew ) {
            my $track_skew = Gdraw2::CairoTrack->new(
                displayName   => 'GC skew track',
                start         => 1,
                end           => $len,
                showGuideLine => 0
            );
            $group->addChild($track_skew);

            my @data;
            my $skew_start;
            my $skew_end;
            for (
                $skew_start = $start ;
                $skew_start < $end ;
                $skew_start += $config->skewWindowStep
              )
            {
                $skew_end = $skew_start + $config->skewWindowSize - 1;
                last if $skew_end > $end;
                push @data,
                  calc_gc_skew( $seq_obj->subseq( $skew_start, $skew_end ) );
            }

            $track_skew->addChild(
                Gdraw2::Element::CairoDataCurve->new(
                    displayName => 'GC skew',
                    position    => [ $start, $skew_end ],
                    middleLine  => 0,
                    data        => [@data]
                )
            );
        }
    }

    # add group to page
    foreach my $group (@groups) {
        my $path = $self->append_group($group);
        my $act  = Gdraw2Gtk::Act::Add->new(
            object  => $group,
            task    => $self,
            indices => [ $path->get_indices ]
        );
        $self->add_act($act);
    }

    # redraw
    my $builder_task = $self->uiBuilder;
    my $area         = $builder_task->get_object('PageComp_PreviewArea');
    $self->page->draw;
    $self->on_area_expose( undef, $area );
}

sub importBlast {
    my ( $self, $config, $IN ) = @_;

    my $do_evalue  = $config->doFiltEValue;
    my $do_score   = $config->doFiltBitScore;
    my $cut_evalue = $config->eValueCutoff;
    my $cut_score  = $config->bitScoreCutoff;

    my $stack_hsp = $config->stackHSPs;

    my @groups;

    while ( my $result = $IN->next_result ) {
        
        #
        # traverse hits
        #
        my @tracks;
        my @elems;

        while ( my $hit = $result->next_hit ) {
            my @hit_elems;

            while ( my $hsp = $hit->next_hsp ) {
                
                # create Block from HSP
                my $score  = $hsp->bits;
                my $evalue = $hsp->evalue;
            
                next if $do_score  and $score < $cut_score;
                next if $do_evalue and $evalue > $cut_evalue;
            
                my $start  = $hsp->start('query');
                my $end    = $hsp->end('query');
                my $strand = $hsp->strand('query') * $hsp->strand('hit');
                my $name   = $hsp->start('hit') . '-' . $hsp->end('hit');
                
                $name = $hit->name .': '.$name if $stack_hsp;
            
                my @color;
                if    ( $score < 40 )  { @color = ( 0, 0, 0, 1 ) }
                elsif ( $score < 50 )  { @color = ( 0, 0, 1, 1 ) }
                elsif ( $score < 80 )  { @color = ( 0, 1, 0, 1 ) }
                elsif ( $score < 200 ) { @color = ( 1, 0, 1, 1 ) }
                else                   { @color = ( 1, 0, 0, 1 ) }
            
                push @hit_elems, Gdraw2::Element::CairoBlock->new(
                    displayName => $name,
                    showName    => 0,
                    position    => [ $start, $end ],
                    strand      => $strand,
                    strokeColor => Gdraw2::Util::RGBAColor->new(@color),
                    fillColor   => Gdraw2::Util::RGBAColor->new(@color)
                );
            }

            if ( @hit_elems > 0 ) {
                if ($stack_hsp) {
                    push @elems, @hit_elems;
                }
                else {
                    my $track = Gdraw2::CairoTrack->new(
                        displayName  => $hit->name,
                        start        => 1,
                        end          => $result->query_length,
                        elementWidth => 5
                    );
                    push @tracks, $track;
                    $track->addChild($_) foreach @hit_elems;
                }
            }
        }
        
        #
        # if non-empty, create the group and ruler
        #
        my $group;

        if ( ( $stack_hsp and @elems > 0 ) or ( !$stack_hsp and @tracks > 0 ) )
        {
            # group
            $group = Gdraw2::CairoTrackGroup->new(
                displayName   => $result->query_name,
                desc          => $result->query_description,
                trackDistance => 0,
                start         => 1,
                end           => $result->query_length
            );
            push @groups, $group;

            # ruler
            my $track_ruler = Gdraw2::CairoTrack->new(
                displayName   => 'ruler track',
                start         => 1,
                end           => $result->query_length,
                showGuideLine => 0
            );
            $group->addChild($track_ruler);

            $track_ruler->addChild(
                Gdraw2::Element::CairoRuler->new(
                    displayName => 'ruler',
                    position    => [ 1, $result->query_length ],
                    gridLength  => calc_tick_size( $result->query_length )
                )
            );
        }
        
        #
        # if non-empty, add contents to group
        #
        if ( defined $group ) {
            if ($stack_hsp) {
                my $track_content = Gdraw2::CairoTrack->new(
                    displayName   => 'HSPs track',
                    start         => 1,
                    elementWidth  => 5,
                    end           => $result->query_length,
                    showGuideLine => 0
                );
                $group->addChild($track_content);
                
                $track_content->addChild($_) foreach @elems;
            }
            else {
                $group->addChild($_) foreach @tracks;
            }
        }
    }

    # add group to page
    foreach my $group (@groups) {
        my $path = $self->append_group($group);
        my $act  = Gdraw2Gtk::Act::Add->new(
            object  => $group,
            task    => $self,
            indices => [ $path->get_indices ]
        );
        $self->add_act($act);
    }

    # redraw
    my $builder_task = $self->uiBuilder;
    my $area         = $builder_task->get_object('PageComp_PreviewArea');
    $self->page->draw;
    $self->on_area_expose( undef, $area );
}

sub export {
    my ($self,$file,$format) = @_;
    
    $file .= ".$format" if $file !~ /\.$format$/;
    
    say "export to '$file' in $format format";
    
    my $page = $self->page;

    # store old context
    my $cr_old = $page->cairoContext;
    
    # create output surface
    my $surface_out;
    if ($format eq 'png') {
        $surface_out =
          Cairo::ImageSurface->create( 'argb32', $page->width, $page->height );
    }
    elsif ($format eq 'svg') {
        $surface_out =
          Cairo::SvgSurface->create( $file, $page->width, $page->height );
    }
    elsif ($format eq 'pdf') {
        $surface_out =
          Cairo::PdfSurface->create( $file, $page->width, $page->height );
    }
    elsif ($format eq 'ps') {
        $surface_out =
          Cairo::PsSurface->create( $file, $page->width, $page->height );
    }
    
    # write output
    my $cr_out = Cairo::Context->create($surface_out);
    $page->cairoContext($cr_out);
    $page->draw;
    $surface_out->write_to_png($file) if $format eq 'png';
    
    # restore old context
    $page->cairoContext($cr_old);
}

sub append_group {
    my $self  = shift;
    my $group = shift;

    my $page = $self->page;

    # add group to page
    $page->addChild($group);

    # update treeview
    my $tree_model = $self->get_tree_model;
    my $iter_group = $tree_model->append(undef);
    my $path_group = $tree_model->get_path($iter_group);
    my @grp_i      = $path_group->get_indices;
    $self->update_table_at( $grp_i[0] );

    # insert children
    foreach my $track ( $group->getChildren ) {
        my $iter_track = $tree_model->append($iter_group);
        my @i_track    = $tree_model->get_path($iter_track)->get_indices;
        $self->update_table_at(@i_track);

        foreach my $ele ( $track->getChildren ) {
            my $iter_ele = $tree_model->append($iter_track);
            my @i_ele    = $tree_model->get_path($iter_ele)->get_indices;
            $self->update_table_at(@i_ele);
        }
    }

    # select at newly added group

    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path_group);
    $treeview->set_cursor($path_group);

    return $path_group;
}

sub add_group_at {
    my $self  = shift;
    my $group = shift;
    my $grp_i = shift;

    # add group to page
    $self->page->addChildAt( $group, $grp_i );

    # update treeview
    my $tree_model = $self->get_tree_model;
    my $iter_group = $tree_model->insert( undef, $grp_i );
    $self->update_table_at($grp_i);

    # insert children
    foreach my $track ( $group->getChildren ) {
        my $iter_track = $tree_model->append($iter_group);
        my @i_track    = $tree_model->get_path($iter_track)->get_indices;
        $self->update_table_at(@i_track);

        foreach my $ele ( $track->getChildren ) {
            my $iter_ele = $tree_model->append($iter_track);
            my @i_ele    = $tree_model->get_path($iter_ele)->get_indices;
            $self->update_table_at(@i_ele);
        }
    }

    # select at newly added group
    my $path     = $tree_model->get_path($iter_group);
    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path);
    $treeview->set_cursor($path);

    return $path;
}

sub remove_group {
    my $self = shift;
    my $grp  = shift;
    return $self->remove_group_at( $self->page->childIndex($grp) );
}

sub remove_group_at {
    my $self  = shift;
    my $grp_i = shift;

    # remove group from page
    my $page  = $self->page;
    my $group = $page->removeChildAt($grp_i);
    confess "" if !defined $group;

    # update_ui
    my $tree_model = $self->get_tree_model;
    my $iter       = $tree_model->get_iter_from_string($grp_i);
    $tree_model->remove($iter);

    # return removed
    return $group;
}

sub append_track {
    my $self      = shift;
    my $track     = shift;
    my $grp_index = shift;

    # get parent
    my $page  = $self->page;
    my $group = $page->getChildAt($grp_index);

    # append track to group
    $group->addChild($track);

    # get iter pointing to end
    my $tree_model = $self->get_tree_model;
    my $iter_group = $tree_model->get_iter_from_string($grp_index);
    my $iter_track = $tree_model->append($iter_group);

    # update UI
    my $path_track = $tree_model->get_path($iter_track);
    my @i_track    = $path_track->get_indices;
    $self->update_table_at(@i_track);

    foreach my $ele ( $track->getChildren ) {
        my $iter_ele = $tree_model->append($iter_track);
        my @i_ele    = $tree_model->get_path($iter_ele)->get_indices;
        $self->update_table_at(@i_ele);
    }

    # select at newly added group
    my $path     = $tree_model->get_path($iter_track);
    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path);
    $treeview->set_cursor($path);

    return $path;
}

sub add_track_at {
    my $self  = shift;
    my $track = shift;
    my $grp_i = shift;
    my $trk_i = shift;

    # get parent
    my $page  = $self->page;
    my $group = $page->getChildAt($grp_i);

    # add to group
    $group->addChildAt( $track, $trk_i );

    # get iter
    my $tree_model = $self->get_tree_model;
    my $iter_group = $tree_model->get_iter_from_string($grp_i);
    my $iter       = $tree_model->insert( $iter_group, $trk_i );

    # update treeview
    $self->update_table_at( $grp_i, $trk_i );

    foreach my $ele ( $track->getChildren ) {
        my $iter_ele = $tree_model->append($iter);
        my @i_ele    = $tree_model->get_path($iter_ele)->get_indices;
        $self->update_table_at(@i_ele);
    }

    # select at newly added group
    my $path     = $tree_model->get_path($iter);
    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path);
    $treeview->set_cursor($path);

    return $path;
}

sub remove_track_at {
    my $self  = shift;
    my $grp_i = shift;
    my $trk_i = shift;

    # get parent
    my $page = $self->page;
    my $grp  = $page->getChildAt($grp_i);

    # remove track from group
    my $track = $grp->removeChildAt($trk_i);

    # update UI
    my $tree_model = $self->get_tree_model;
    my $iter = $tree_model->get_iter_from_string( join ':', $grp_i, $trk_i );
    $tree_model->remove($iter);

    # return removed
    return $track;
}

sub append_element {
    my $self  = shift;
    my $ele   = shift;
    my $grp_i = shift;
    my $trk_i = shift;

    # get parent
    my $page  = $self->page;
    my $group = $page->getChildAt($grp_i);
    my $track = $group->getChildAt($trk_i);

    # add element to track
    $track->addChild($ele);

    # get iter pointing to end
    my $tree_model = $self->get_tree_model;
    my $iter_track = $tree_model->get_iter_from_string( $grp_i . ':' . $trk_i );
    my $iter       = $tree_model->append($iter_track);

    # update UI
    $self->update_table_at( $grp_i, $trk_i, $track->numChildren - 1 );

    # select at newly added group
    my $path     = $tree_model->get_path($iter);
    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path);
    $treeview->set_cursor($path);

    return $path;
}

sub add_element_at {
    my ( $self, $ele, $grp_i, $trk_i, $ele_i ) = @_;

    # get parent
    my $page  = $self->page;
    my $group = $page->getChildAt($grp_i);
    my $track = $group->getChildAt($trk_i);

    # add to parent
    $track->addChildAt( $ele, $ele_i );

    # add to table
    my $tree_model = $self->get_tree_model;
    my $iter_track =
      $tree_model->get_iter_from_string( join ':', $grp_i, $trk_i );
    my $iter = $tree_model->insert( $iter_track, $ele_i );

    # update UI
    $self->update_table_at( $grp_i, $trk_i, $ele_i );

    # select at newly added group
    my $path     = $tree_model->get_path($iter);
    my $treeview = $self->get_tree_view;
    $treeview->expand_to_path($path);
    $treeview->set_cursor($path);

    return $path;
}

sub remove_element_at {
    my $self  = shift;
    my $grp_i = shift;
    my $trk_i = shift;
    my $ele_i = shift;

    # get parent
    my $page = $self->page;
    my $grp  = $page->getChildAt($grp_i);
    my $trk  = $grp->getChildAt($trk_i);

    # remove element
    my $ele = $trk->removeChildAt($ele_i);

    # update treeview UI
    my $tree_model = $self->get_tree_model;
    my $iter =
      $tree_model->get_iter_from_string( join ':', $grp_i, $trk_i, $ele_i );
    $tree_model->remove($iter);

    # return deleted
    return $ele;
}

sub update_table_at {
    my ( $self, @i ) = @_;

    my $tree_model = $self->get_tree_model;
    my $iter = $tree_model->get_iter_from_string( join ':', @i );

    my $page = $self->page;

    if ( @i == 3 ) {
        my $ele =
          $page->getChildAt( $i[0] )->getChildAt( $i[1] )->getChildAt( $i[2] );
        $tree_model->set(
            $iter,                           COL_NAME,
            $ele->displayName,               COL_LOCATION,
            $ele->start . '..' . $ele->end,  COL_STROKE,
            $ele->strokeColor->toGdkString8, COL_FILL,
            $ele->fillColor->toGdkString8,   COL_COLOR_SET,
            TRUE
        );
    }
    elsif ( @i == 2 ) {
        my $track = $page->getChildAt( $i[0] )->getChildAt( $i[1] );
        $tree_model->set( $iter, COL_NAME, $track->displayName, COL_LOCATION,
            $track->start . '..' . $track->end,
            COL_COLOR_SET, FALSE );
    }
    elsif ( @i == 1 ) {
        my $group = $page->getChildAt( $i[0] );
        $tree_model->set( $iter, COL_NAME, $group->displayName, COL_LOCATION,
            $group->start . '..' . $group->end,
            COL_COLOR_SET, FALSE );
    }
    else {
        confess "invalid indices: @i";
    }
}

sub rebuild_treeview {
    my $self          = shift;
    my $selection_str = shift;

    # remove everything
    my $tree_model = $self->get_tree_model;
    while (1) {
        my $iter_first = $tree_model->get_iter_first;
        last if !defined $iter_first;
        $tree_model->remove($iter_first);
    }

    # re-generate everything
    foreach my $group ( $self->page->getChildren ) {
        my $group_iter = $tree_model->append(undef);
        my @group_i    = $tree_model->get_path($group_iter)->get_indices;
        $self->update_table_at(@group_i);

        foreach my $track ( $group->getChildren ) {
            my $track_iter = $tree_model->append($group_iter);
            my @track_i    = $tree_model->get_path($track_iter)->get_indices;
            $self->update_table_at(@track_i);

            foreach my $ele ( $track->getChildren ) {
                my $ele_iter = $tree_model->append($track_iter);
                my @ele_i    = $tree_model->get_path($ele_iter)->get_indices;
                $self->update_table_at(@ele_i);
            }
        }
    }

    # set cursor if given
    if ( defined $selection_str and $selection_str ne '' ) {
        my $path     = Gtk2::TreePath->new_from_string($selection_str);
        my $treeview = $self->get_tree_view;
        $treeview->expand_to_path($path);
        $treeview->set_cursor($path);
    }
}

sub undo {
    my $self = shift;
    my $re   = $self->can_undo;
    return 0 if !$re;

    my $act = pop @{ $self->undo_list };
    $act->undo;
    push @{ $self->redo_list }, $act;
    return 1;
}

sub redo {
    my $self = shift;
    return 0 if !$self->can_redo;

    my $act = pop @{ $self->redo_list };
    $act->redo;
    push @{ $self->undo_list }, $act;
    return 1;
}

sub can_undo {
    my $self = shift;
    return ( @{ $self->undo_list } > 0 );
}

sub can_redo {
    my $self = shift;
    return ( @{ $self->redo_list } > 0 );
}

sub add_act {
    my ( $self, $act ) = @_;
    $self->{redo_list} = [];
    push @{ $self->undo_list }, $act;
}

sub on_area_expose {
    my ( $self, $event, $area ) = @_;

    my $surface_buffer = $self->page->cairoContext->get_target;
    my $cr_area        = Gtk2::Gdk::Cairo::Context->create( $area->window );
    $cr_area->set_source_surface( $surface_buffer, 0, 0 );
    $cr_area->paint;
}
1;

